// Author: Juraj Vanek
// This shader handles extraction of bright parts of the scene and computation of SSAO effect
// SSAO code adapted from http://www.gamerendering.com/2009/01/14/ssao/

//texture samplers
in vec2 fragTexCoord;
uniform sampler2D renderTexture, normalTexture, mat_bloomBaseA;
//output color
out vec4 out_FragColor;

#ifdef BLOOM
  //Bloom variables
  uniform float THRESHOLD;
#endif

#ifdef SSAO

  //SSAO random vectors sphere
  const vec3 pSphere[8] = vec3[8](
      vec3(0.24710192, 0.6445882, 0.033550154),
      vec3(0.00991752, -0.21947019, 0.7196721),
      vec3(0.25109035, -0.1787317, -0.011580509),
      vec3(-0.08781511, 0.44514698, 0.56647956),
      vec3(-0.011737816, -0.0643377, 0.16030222),
      vec3(0.035941467, 0.04990871, -0.46533614),
      vec3(-0.058801126, 0.7347013, -0.25399926),
      vec3(-0.24799341, -0.022052078, -0.13399573)
  );

  //SSAO control variables
  uniform float totStrength, rad, bias;

  const int SAMPLES = 8;
  const float invSamples = 1.0/8.0;
  const float offset = 50.0;

  //Calculates SSAO term using normal texture with depth factor as alpha value
  float ssao(sampler2D normalTex, sampler2D randomTex, vec2 uv)
  {
      // grab a normal for reflecting the sample rays later on
     vec3 fres = normalize((texture2D(randomTex, uv*offset).xyz*2.0) - vec3(1.0));
     vec4 currentPixelSample = texture2D(normalTex,uv);
     float currentPixelDepth = currentPixelSample.a;

     // current fragment coords in screen space
     vec3 ep = vec3(uv.xy,currentPixelDepth);
      // get the normal of current fragment
     vec3 norm = currentPixelSample.xyz;

     float bl = 0.0;
     // adjust for the depth
     float radD = rad/currentPixelDepth;

     vec3 ray, se, occNorm;
     float occluderDepth, depthDifference, normDiff;

     for(int i=0; i<SAMPLES;++i)
     {
        // get a vector (randomized inside of a sphere with radius 1.0) from a texture and reflect it
        ray = radD*reflect(pSphere[i],fres);
        // if the ray is outside the hemisphere then change direction
        se = ep + sign(dot(ray,norm) )*ray;
        // get the depth of the occluder fragment
        vec4 occluderFragment = texture2D(normalTex,se.xy);
        // get the normal of the occluder fragment
        occNorm = occluderFragment.xyz;
        // if depthDifference is negative = occluder is behind current fragment
        depthDifference = currentPixelDepth - occluderFragment.a;
        // calculate the difference between the normals as a weight
        normDiff = (1.0-dot(occNorm,norm));
        // the falloff equation, starts at falloff and is kind of 1/x^2 falling
        bl += step(0.0,depthDifference)*normDiff*(1.0-smoothstep(0.0,bias,depthDifference));
     }
     // output the result
     return 1.0-totStrength*bl*invSamples;
  }
#endif

//////
//MAIN
void main()
{
  vec3 result = vec3(0.0);
  float ssao_term = 1.0;
  int i,j;
  
#ifdef BLOOM
  //extract brightest parts of image
  vec4 texel = texture2D(renderTexture,fragTexCoord);
  if( dot(vec3(0.30, 0.59, 0.11),texel.rgb) > THRESHOLD)
    result = texel.rgb;
    
#endif

#ifdef SSAO
  //compute SSAO term
  ssao_term = ssao(normalTexture, mat_bloomBaseA, fragTexCoord);
#endif

  out_FragColor.rgb = result;
  out_FragColor.a = ssao_term;
}
