//code adapted from: http://web.engr.oregonstate.edu/~mjb/cs519/Handouts/geometry_shader.pdf
#extension GL_ARB_geometry_shader4: enable
#extension GL_EXT_gpu_shader4: enable

layout(triangles) in;
layout(max_vertices=255) out;

uniform sampler2D mat_tessBumpA, mat_tessBaseA;
uniform int TESS_LEVEL;

//computed vertices and texcoords
vec4 V0, V01, V02;
vec2 T0, T01, T02;

//material settings
struct Material{
  vec3 ambient;
  vec3 diffuse;
  vec3 specular;
  float shininess;
};
uniform Material material;

in vec2 geomTexCoord[];

out vec4 color;
out float fogFactor;
out vec2 fragTexCoord;


#define FOG_START 10.0
#define FOG_END 200.0

void ProduceVertex( float s, float t )
{
    vec4 vertex = V0 + s*V01 + t*V02;
    vec2 coord = T0 + s*T01 + t*T02;

    //displace vertex and two neighbouring vertices
    float step = 0.01;
    vec4 dv = texture2D( mat_tessBumpA, vec2(coord.s, 3.0*coord.t) )
    //TODO: WTF??? proc musim samplovat aj Base mapy len aby to islo aj v fragment shaderi???
        + 0.0*texture2D( mat_tessBaseA, vec2(0.0,0.0) ) ;
    float displace = dot(vec3(0.30,0.59,0.11),dv.rgb);
    //second vertex
    vec4 vertex1 = vertex;
    vec4 dv1 = texture2D( mat_tessBumpA, vec2(coord.s + step, 3.0*coord.t) );
    float displace1 = dot(vec3(0.30,0.59,0.11),dv1.rgb);
    vertex1.x += 1.0;
    vertex1.y += 20.0*displace1;
    //third vertex
    vec4 vertex2 = vertex;
    vec4 dv2 = texture2D( mat_tessBumpA, vec2(coord.s, 3.0*coord.t + step) );
    float displace2 = dot(vec3(0.30,0.59,0.11),dv2.rgb);
    vertex2.y += 20.0*displace2;
    vertex2.z += 1.0;
    //original vertex
    vertex.y += 20.0*displace;

    //displaced normal
    vec3 normal = vec3(gl_ModelViewMatrix * vec4(normalize(cross(vertex2.xyz - vertex.xyz, vertex1.xyz - vertex.xyz)),0.0));
    
    //eye vector
    vec3 eyeVec = -vec3(gl_ModelViewMatrix * vertex);

    //light model
    color.rgb = gl_LightSource[0].ambient.rgb * material.ambient;
    vec3 N = normalize(normal);
    vec3 L = normalize(gl_LightSource[0].position.xyz);
    color.rgb += gl_LightSource[0].diffuse.rgb * material.diffuse * dot(N,L);

    fragTexCoord = coord;
    //add fog
    fogFactor = (FOG_END - length(eyeVec)) / (FOG_END - FOG_START);
    fogFactor = clamp(fogFactor, 0.0, 1.0);

    gl_Position = gl_ModelViewProjectionMatrix * vertex;
    EmitVertex();
}

void main()
{
    //starting points for s,t transformation V(s,t) = V0 + s*(V1-V0) + t*(V2-V0)
    V01 = gl_PositionIn[1] - gl_PositionIn[0];
    V02 = gl_PositionIn[2] - gl_PositionIn[0];
    V0 = gl_PositionIn[0];
    //starting points for texture coordinates
    T01 = geomTexCoord[1] - geomTexCoord[0]; 
    T02 = geomTexCoord[2] - geomTexCoord[0];
    T0 = geomTexCoord[0];

    int level = TESS_LEVEL;

    //distances from vertex to camera for every input vertex
    float dist1 = length(vec3(gl_ModelViewMatrix * gl_PositionIn[01]));
    float dist2 = length(vec3(gl_ModelViewMatrix * gl_PositionIn[1]));
    float dist3 = length(vec3(gl_ModelViewMatrix * gl_PositionIn[2]));
    //find out the greatest
    float dist = min(min(dist1,dist2),dist3);

    //LOD: lower tesselation level for primitives too far from camera
    if( dist > 100.0)
        level /= 4;
    else if(dist > 75.0 && dist < 100.0)
        level /= 2;

    int numLayers = 1 << level;             //number of subdivisions on each edge of triangle (2^level)
    float dt = 1.0 / float( numLayers );    //step
    float t_top = 1.0;                      //initial relative position

    for( int it = 0; it < numLayers; it++ )
    {
        //move to the next point
        float t_bot = t_top - dt;
        float smax_top = 1.0 - t_top;
        float smax_bot = 1.0 - t_bot;
        int nums = it + 1;

        float ds_top = smax_top / float( nums - 1 );    //move increment on s-side (top)
        float ds_bot = smax_bot / float( nums );        //move increment on s-side (bottom)
        float s_top = 0.0;
        float s_bot = 0.0;
        //generate vertices
        for( int is = 0; is < nums; is++ )
        {
            ProduceVertex( s_bot, t_bot );
            ProduceVertex( s_top, t_top );
            s_top += ds_top;
            s_bot += ds_bot;
        }
        //finish new triangle
        ProduceVertex( s_bot, t_bot );
        EndPrimitive();
        t_top = t_bot;
        t_bot -= dt;
    }
}
