/*
//vertex attributes
in vec3 in_Vertex;
in vec3 in_Normal;
in vec2 in_Coord;
*/

//built-in vertex attributes
vec3 in_Vertex = gl_Vertex.xyz;
vec3 in_Normal = gl_Normal;
vec2 in_Coord = gl_MultiTexCoord0.st;

uniform float lightRadius;
uniform vec3 camPos;
uniform float time;
out vec4 v_color;
out vec3 fnormal, feyeVec;
out float fogFactor;
out vec2 fragTexCoord;

//material settings
struct Material{
  vec3 ambient;
  vec3 diffuse;
  vec3 specular;
  float shininess;
};
uniform Material material;

//fog variables
const float FOG_START = 100.0;
const float FOG_END = 400.0;

//Returns random 4-dimensional vector
vec4 rand(vec2 A,vec2 B,vec2 C,vec2 D)
{
    vec2 s = vec2 (12.9898,78.233);
    vec4 tmp = vec4( dot(A,s),dot(B,s),dot(C,s),dot(D,s));

	return fract(sin(tmp)  * 43758.5453);
 }

//Calculate basic linear-filtered noise
float noise(vec2 coord,float d)
{
    vec2 C[4];
    float d1 = 1.0/d;
    C[0] = floor(coord * d)*d1 ;
    C[1] = C[0] + vec2(d1 ,0.0 );
    C[2] = C[0] + vec2(d1 ,d1);
    C[3] = C[0] + vec2(0.0,d1);

    vec2 p = fract(coord * d);
    vec2 q = 1.0 - p;
    vec4 w = vec4(q.x * q.y, p.x * q.y, p.x * p.y, q.x * p.y);

    return dot(rand(C[0],C[1],C[2],C[3]),w);
 }

//Creates 2D Perlin noise
float perlinNoise2D(vec2 coord, float persistence, float oct)
{
	float total = 0.0;

    for(float i=1.0; i<=oct; i++)
    {
        float frequency = pow(2.0,i);
        float amplitude = pow(persistence,i);

        total += noise(coord*frequency+time/10.0, amplitude)/oct;
    }
    return total;
}


///////////////////////////////////////////////////////////////////////////////
//MAIN
void main()
{
  //we need actual and two neigbouring vertices to calculate normal
  float step = 2.0;
  vec4 vertex = vec4(in_Vertex,1.0);
  vec4 vertexX = vertex;
  vertexX.x += step;
  vec4 vertexZ = vertex;
  vertexZ.z += step;
  
  //calculate wave values for all three vertices
  vertex.y = 3.0*sin(vertex.x/16.0 + time) + sin(vertex.x/8.0 + 2.0*time) + sin(vertex.z/10.0 + time) - 3.5*perlinNoise2D(vertex.xz/512.0,2.0,4.0);
  vertexX.y = 3.0*sin(vertexX.x/16.0 + time) + sin(vertexX.x/8.0 + 2.0*time) + sin(vertexX.z/10.0 + time) - 3.5*perlinNoise2D(vertexX.xz/512.0,2.0,4.0);
  vertexZ.y = 3.0*sin(vertexZ.x/16.0 + time) + sin(vertexZ.x/8.0 + 2.0*time) + sin(vertexZ.z/10.0 + time) - 3.5*perlinNoise2D(vertexZ.xz/512.0,2.0,4.0);
  
  //displaced normal
  vec3 dnormal = normalize(cross(vertexZ.xyz - vertex.xyz, vertexX.xyz - vertex.xyz));

  //camera rotation matrix
  mat3 tcamerarot;
  tcamerarot[0][0] = gl_ModelViewMatrixTranspose[0][0];
  tcamerarot[0][1] = gl_ModelViewMatrixTranspose[0][1];
  tcamerarot[0][2] = gl_ModelViewMatrixTranspose[0][2];

  tcamerarot[1][0] = gl_ModelViewMatrixTranspose[1][0];
  tcamerarot[1][1] = gl_ModelViewMatrixTranspose[1][1];
  tcamerarot[1][2] = gl_ModelViewMatrixTranspose[1][2];

  tcamerarot[2][0] = gl_ModelViewMatrixTranspose[2][0];
  tcamerarot[2][1] = gl_ModelViewMatrixTranspose[2][1];
  tcamerarot[2][2] = gl_ModelViewMatrixTranspose[2][2];
    
  vec3 normal = gl_NormalMatrix * dnormal;            //surface normal
  vec3 eyeVec = -vec3(gl_ModelViewMatrix * vertex);  //view vector
  
  //normal and eyevec for cubemapping
  fnormal = tcamerarot * normal;
  feyeVec = tcamerarot * -eyeVec;

  //calculate light model
  vec3 N = normalize(normal);
  vec3 E = normalize(eyeVec);
  v_color.rgb = gl_LightSource[0].ambient.rgb * material.ambient;
  vec3 L = normalize(gl_LightSource[0].position.xyz + eyeVec + camPos);

  float lambertTerm = dot(N,L);
  if(lambertTerm > 0.0)
  {
  	v_color.rgb += gl_LightSource[0].diffuse.rgb * material.diffuse * lambertTerm;
  	vec3 R = reflect(-L, N);
  	float specular = pow( max(dot(R, E), 0.0), material.shininess );
  	v_color.rgb += gl_LightSource[0].specular.rgb *  material.specular * specular;
  }

  //add fog for eye-candy
  fogFactor = clamp((FOG_END - length(eyeVec)) / (FOG_END - FOG_START), 0.0, 1.0);

  fragTexCoord = in_Coord;
  gl_Position = gl_ModelViewProjectionMatrix * vertex;      //vertex transform
}
