/**
****************************************************************************************************
****************************************************************************************************
@file light.h
@brief light class and light settings
****************************************************************************************************
***************************************************************************************************/
#ifndef _LIGHT_H_
#define _LIGHT_H_
#include "globals.h"
#include "mat.h"
#include "object.h"

///@class TLight 
///@brief contains light preferences like color,position,shadow...
class TLight
{
    private:
        TVector pos;        //position
        TVector ambient, diffuse, specular;     //light settings
				float radius;
        GLint light;        //light number
        bool shadow;        //has light shadow?
    
        //shadow parameters
        GLint shadow_size;      //shadow map size
        GLfloat shadow_intensity;  //intensity of shadow

				GLUquadricObj *q;
        
    public:
        ///to calculate shadows, we need framebuffer for shadow map
        GLuint fbo;         
        ///light projection matrix
        TMatrix lightProjectionMatrix;
				///light view matrix
        TMatrix lightViewMatrix;
				///camera projection matrix
        TMatrix cameraProjectionMatrix;
				///camer aview matrix
        TMatrix cameraViewMatrix;
				///shadow bias matrix
        TMatrix biasMatrix;
				///shadow texture matrix
        TMatrix textureMatrix;
        
        TLight();
        ~TLight();
        /**
				@brief Main constructor, directly create light from parameters
				@param lights light index
				@param amb ambient light color
				@param diff diffuse light color
				@param spec specular light color
				@param lpos vector with light position
				@param radius light attenuation radius
				***********************************************************************************/
        TLight(GLint lights, TVector amb, TVector diff, TVector spec, TVector lpos, float radius)
        { Set(lights,amb,diff,spec,lpos,radius); }
        
        ///@brief Move light to new absolute position
        void Move(TVector w) 
				{  pos = w; glLightfv(light + GL_LIGHT0, GL_POSITION, w); }
            
        ///@brief Keep light position after camera transform    
        void KeepPos()         
        {  glLightfv(light + GL_LIGHT0, GL_POSITION, pos); }
    
        void DrawLight(TMatrix &transformMatrix);       //Draws current light position
            
        //change color
        void ChangeColor(GLint component, TVector color);
        //set all light settings
        void Set(GLint lights, TVector amb, TVector diff, TVector spec, TVector lpos, float radius);
        //sets light shadow
        void SetShadow(GLint shadow_size, GLfloat _shadow_intensity, bool shadow);
				///@brief set light radius
				void SetRadius(float value)
				{ radius = value; }

        ///@brief Returns light shadow
        bool HasShadow() 
				{ return shadow; } 
        ///@brief return shadow map size
        GLint ShadowSize() 
				{ return shadow_size; }  
        ///@brief return shadow intensity
        GLfloat ShadowIntensity() 
				{ return shadow_intensity; } 
        //get shadow cache
        string GetShadowCache();       


        ///@brief Return light position
        TVector GetPos()
        { return pos; }
        ///@brief Get light number
        GLint GetOrd() 
        { return light; }
				///@brief Get light radius
				float GetRadius()
				{ return radius; }

};

#endif
