/**
@file material.h 
@brief material settings, dynamic shader creation
***************************************************************************************************/
#ifndef _MATERIAL_H_
#define _MATERIAL_H_

#include "texture.h"

///Aligned buffer size
#define BUFFER 512

///white color
const TVector white(1.0f,1.0f,1.0f);
///black color
const TVector black(0.0f,0.0f,0.0f);
///dark grey color
const TVector dgrey(0.1f,0.1f,0.1f);
///grey color
const TVector grey(0.3f,0.3f,0.3f);
///light grey color
const TVector lgrey(0.5f,0.5f,0.5f);
///silver color
const TVector silver(0.7f,0.7f,0.7f);
///red color
const TVector red(1.0f,0.0f,0.0f);
///green color
const TVector green(0.0f,1.0f,0.0f);
///blue color
const TVector blue(0.0f,0.0f,1.0f);
///yellow color
const TVector yellow(1.0f,1.0f,0.0f);
///magenta color
const TVector magenta(1.0f,0.0f,1.0f);
///cyan color
const TVector cyan(0.0f,1.0f,1.0f);


///@class TMaterial
///@brief hold all material properties necessary to create dynamic shader - light models textures and colors
class TMaterial
{
    private:
        //properties
        string name;
        map<string,Texture> textures;          //textures list    
        map<string,Texture>::iterator it;			 //texture iterator
				//material properties
        TVector ambColor, diffColor, specColor;
        float shininess, transparency, reflection;

        //shader
        string source;          //custom shader source
        GLuint f_shader, g_shader, v_shader, shader, gshader_ref;

        //other variables
        bool baked, custom_shader, receive_shadows, use_gshader_ref, useMRT;
        int lightModel;     

				//scene ID - when drawing more scenes than 1
				int sceneID;

    public:
				//default material
        TMaterial()
				{	TMaterial("basic",black,silver,white,64.0,0.0,0.0,NONE); }
        //full material creation
        TMaterial(const char* _name, TVector amb, TVector diff, TVector spec, GLfloat shin, GLfloat reflect, GLfloat transp, GLint lm);
				//free dynamic data
        ~TMaterial();
        
        ///@brief get material name
        string GetName() 
				{ return name; }
        ///@brief get transparency value
        GLfloat GetTrasparency() 
				{ return transparency; }


        //texture from file
        GLint AddTexture(const char *file, GLint textype, GLint texmode,
                         GLfloat intensity, GLfloat tileX, GLfloat tileY, bool mipmap, bool aniso, GLint cache);
        //from cubemap
        GLint AddTexture(const char **files, GLint textype, GLint texmode,
                         GLfloat intensity, GLfloat tileX, GLfloat tileY, bool aniso, GLint cache);
         //finds next free texture in the list
        string NextTexture(string textype);   
				/**
				@brief Delete texture specified by name
				@param texname texture name
				***************************************/
				void DeleteTexture(string texname)
				{ textures.erase(texname); }


        //add shadow map
        void AddShadowMap(GLuint map, GLfloat intensity);
        //set/update shadow matrix to shader
        void SetShadowMatrix(TMatrix &m, GLint l_num);
				//remove all shadow maps
				void RemoveShadows();
   

        //load custom shader
        bool CustomShader(const char* vert_source, const char* geom_source, const char* frag_source, 
                          const char *vert_defines, const char *frag_defines, const char *geom_defines);
		    //set uniform values in shader
		    void SetUniform(const char* v_name, float value);     //float value
        void SetUniform(const char* v_name, double value);    //double value is treated like float
        void SetUniform(const char* v_name, int value);       //integer values
        //set array uniform variable (eg. vector)
        void SetUniforms(const char* v_name, float *value, int values);
        ///@brief Toggle use of geometry shader in dynamic cubemapping
        void UseGShaderCubeMapRender(bool flag)
        { use_gshader_ref = flag; }
        ///@brief Toggle use of MRT
        void UseMRT(bool flag)
        { useMRT = flag; }

        //dynamically generate material shader
        bool BakeMaterial(bool use_geom = false);
        //render material
        void RenderMaterial(bool render_all = true);
        ///@brief set number of lights to use in shader
        void SetLights(int lights) 
        { SetUniform("LIGHTS",lights); }

        ///@brief toggle receiving shadows
        void ReceiveShadow(bool flag) 
        { receive_shadows = flag; }

        ///@brief should we render material as reflective?
        float GetReflect() 
        { return reflection; }

				///Get scene ID
				int GetSceneID()
				{  return sceneID; }
				///Set scene ID
				void SetSceneID(int id)
				{  sceneID = id; }
};

#endif
