/**
@file material_generator.cpp
@brief dynamic shader creation
***************************************************************************************************/
#include "material.h"


/**
****************************************************************************************************
@brief Load shader function from file
@param func shader function source file 
@return string with shader function source
***************************************************************************************************/
string LoadFunc(char* func)
{
  string file;
  file += "data/shaders/func/";
  file += func;
  file += ".shad";
  ifstream fin(file.c_str());
  if(!fin) 
		return "null";
  string data;
  char ch;
  while(fin.get(ch))
    data+=ch;

  return data;
}


////////////////////////////////////////////////////////////////////////////////
//************************* TMaterial methods ********************************//
////////////////////////////////////////////////////////////////////////////////

/**
****************************************************************************************************
@brief Computes texel color accordig to texture mode
@param t pointer to texture data
@param name texture name
@return shader piece of code
***************************************************************************************************/
string computeTexel(Texture &t, string name)
{
  string ret = "\n";

  //environment mapping - we need function to generate reflection coordinates
  if(t.GetType() == ENV) 
		ret += "  " + name + "_texture = envMapping(color,eyeVec,normal," + name + ", " + name + "_intensity);\n";
  //cube map - use samplerCube
  else if(t.GetType() == CUBEMAP) 
		ret += "  " + name + "_texture = textureCube(" + name + ", " + name + "_cubeCoords);\n";
  //cube map - use reflection vector
  else if(t.GetType() == CUBEMAP_ENV) 
		ret += "  " + name + "_texture = textureCube(" + name + ", reflVec);\n";
  //default 2D texture
  else 
		ret += "  " + name + "_texture = texture2D(" + name + ", " + name + "_texcoord);\n";

  switch(t.GetMode())
  {
    ///ADD mode: add texel RGB values to color RGB values. Alpha channels are multiplied
    case ADD:
      ret +=
        "  //" + name + ", ADD mode\n"
        "  color.rgb += " + name + "_texture.rgb;\n"
        "  color.a *= " + name + "_texture.a;\n"
        "  color = clamp (color, 0.0, 1.0);\n";
        break;
    ///MODULATE mode: multiply texel channels with actual color channels
    case MODULATE:
      ret +=
        "  //" + name + ", MODULATE mode\n"
        "  color *= " + name + "_texture;\n";
        break;
    ///DECAL mode: mix texel and color together
    case DECAL:
      ret +=
        "  //" + name + ", DECAL mode\n"
        "  color = vec4(mix(color.rgb, " + name + "_texture.rgb, " + name + "_texture.a), color.a);\n";
        break;
     ///BLEND mode: mix texel and color together, multiply alpha channels
    case BLEND:
      ret +=
        "  //" + name + ", BLEND mode\n"
        "  color = vec4 ( vec3(mix(color.rgb, gl_TextureEnvColor[0].rgb, " + name + "_texture.rgb)), color.a * " + name + "_texture.a);\n";
        break;
    ///REPLACE mode: replace fragment color directly with texel color
    case REPLACE:
      ret +=
        "  //" + name + ", REPLACE mode\n"
        "  color = " + name + "_texture;\n";
        break;          
  };
  return ret;
}

/**
****************************************************************************************************
@brief Dynamically generates shader from all material data (if hasn't been generated before)
@param use_geom should we generate geometry shader?
@return success/fail of shader generation
***************************************************************************************************/
bool TMaterial::BakeMaterial(bool use_geom)
{
	//dont't bake linked custom shader
	if(custom_shader)
		return true;

	cout<<"Baking material "<<name<<endl;

  string tmp = name;        //temporary string for comparison

  //if geometry shader is present, rename varyings by add v_ prefix so they can be sent through pipeline
  string g_pref;
  if(use_geom) 
	g_pref = "v_";

  /////////////////////////////////////////////////////////////////////////////
  ///1 CREATION OF VERTEX SHADER
  /////////////////////////////////////////////////////////////////////////////

  ///1.1 Vertex shader variables
  string vert_vars = "//GLSL vertex shader generated by gluxEngine\n";    //vertex shader variables
  vert_vars += "//Material: \"" + name + "\"\n\n";
#ifdef GL3
  vert_vars +=
	    "//generic vertex attributes\n"
		"in vec3 in_Vertex;\n"
		"in vec3 in_Normal;\n"
		"in vec2 in_Coord;\n\n";
#else
  vert_vars +=
	    "//built-in vertex attributes\n"
		"vec3 in_Vertex = gl_Vertex.xyz;\n"
		"vec3 in_Normal = gl_Normal;\n"
		"vec2 in_Coord = gl_MultiTexCoord0.st;\n\n";
#endif
  vert_vars += "//texture coordinate\n";
	//different coordinate name for geometry shader
	if(use_geom)
		vert_vars += "out vec2 geomTexCoord;\n";
	else
		vert_vars += "out vec2 fragTexCoord;\n";

  string vert_func;                                                       //vertex shader functions
	//vertex shader main function
  string vert_main = 
		"\nvoid main()\n"
	  "{\n"
	  "  vec4 vertex = vec4(in_Vertex,1.0);\n"
	  "  vec3 dNormal = in_Normal;\n";	//vertices and normals for further calculations 
	
  ///1.1.1 if we have displacement texture, modify vertex position and normal
  bool displace = false;
  for(it = textures.begin(); it != textures.end(); it++)
  {
    if(it->second.GetType() == DISPLACE)
		{
			string texname = it->second.GetName();
			//add texture samplers: for normal and displacement map (assume that normal map is present)
			if(it->second.HasTiles())
				vert_vars += "uniform float " + texname + "_tileX, " + texname + "_tileY;\n";
			vert_vars += "uniform float " + texname + "_intensity;\n"
								   "uniform sampler2D "+ texname + ";\n";
			vert_main +=  "  //displace vertex\n";
			//tiles: tile texture coordinate
			if(it->second.HasTiles())
				 vert_main += "  vec4 dv = texture2DLod(" + texname + ", in_Coord*vec2(" + texname + "_tileX, " + texname + "_tileY), 0.0);\n";
			//no tiles
			else
				 vert_main += "  vec4 dv = texture2DLod(" + texname + ", in_Coord, 0.0);\n";
			//displace vertex along normal
			vert_main += "  float displace = dv.r; //dot( vec4(0.30,0.59,0.11,0.0),dv);\n"
									 "  vertex.xyz += " + texname + "_intensity * displace * in_Normal;\n";
			displace = true;
			break;
		}
	}
	//require also normal map
	if(displace)
	{
		for(it = textures.begin(); it != textures.end(); it++)
		{
			if(it->second.GetType() == BUMP)
			{
				string texname = it->second.GetName();
				if(it->second.HasTiles())
					vert_vars += "uniform float " + texname + "_tileX, " + texname + "_tileY;\n";
				vert_vars += "uniform float " + texname + "_intensity;\n"
										 "uniform sampler2D "+ texname + ";\n";
				vert_main	+= "  //displace normal using normal map\n";
				//tiles: tile texture coordinate
				if(it->second.HasTiles())
					vert_main	+= "  dNormal += normalize(texture2D(" + texname + ",in_Coord*vec2(" + texname + "_tileX, " + texname + "_tileY)).rgb * 2.0 - 1.0);\n\n";
				//no tiles
				else
					vert_main	+= "  dNormal += normalize(texture2D(" + texname + ",in_Coord).rgb * 2.0 - 1.0);\n\n";
				break;
			}
		}
	}


  //*****************************
  ///1.2 Light model: PHONG or GOURAUD

  tmp += "Env";
  ///1.2.1 if PHONG, light calculation will be done in fragment shader, so send necessary variables to it (also required for env mapping)
  if(lightModel == PHONG || textures.find(tmp) != textures.end() )
  {
    vert_vars += "out vec3 " + g_pref + "normal, " + g_pref + "eyeVec;\n";  //varying variables

    vert_main +=
      "  " + g_pref + "normal = gl_NormalMatrix * dNormal;                //surface normal\n"
			"  vec3 vVertex = vec3(gl_ModelViewMatrix * vertex); //view vector\n"
	    "  " + g_pref + "eyeVec = -vVertex;                                   //eyeview position\n";
  }

  ///1.2.3 if GOURAUD, done light calculation per-vertex
  else if(lightModel == GOURAUD)
  {
    ///@todo OpenGL built-in uniforms can't handle more than 8 lights 
    vert_vars += "out vec4 " + g_pref + "v_color;\n" + 
				 "out vec3 " + g_pref + "normal," + g_pref + "eyeVec;\n";

    vert_func += LoadFunc("light");
    vert_main +=
      "  " + g_pref + "normal = gl_NormalMatrix * dNormal;            //surface normal\n"
			"  vec3 vVertex = vec3(gl_ModelViewMatrix * vertex);  //view vector\n"
      "  " + g_pref + "eyeVec = -vVertex;                               //eyeview position\n"
      "  " + g_pref + "v_color = LightModel(" + g_pref + "normal," + g_pref + "eyeVec);         //calculate light model\n";
  }
  //no light model
  else
  {
     vert_vars += "out vec3 normal;\n";
     vert_main += "  normal = gl_NormalMatrix * dNormal;            //surface normal\n";
  }

  //**********************************
  //Textures

	///1.3 send texture coords to fragment shader (or geometry shader)
	if(use_geom)
		vert_main += "  geomTexCoord = in_Coord;\n";
	else
		vert_main += "  fragTexCoord = in_Coord;\n";


  //******************************
  //iterate through textures
  for(it = textures.begin(); it != textures.end(); it++)
  {
    ///1.4 calculate shadow matrices for frag. shader (projected shadow and texture matrix)
    if(it->second.GetType() == SHADOW)
    {
      vert_vars += "out vec4 " + g_pref + it->first + "_projShadow;\n";
      vert_vars += "uniform mat4 " + it->first + "_texMatrix;\n";
      vert_main += "  " + g_pref + it->first + "_projShadow = " + it->first + "_texMatrix * gl_ModelViewMatrix * vertex;   //calculate shadow texture projection\n";
    }

    ///1.5 send 3D coordinates for cube map
    if(it->second.GetType() == CUBEMAP)
    {
       vert_vars += "out vec3 " + g_pref + it->first + "_cubeCoords;\n";
       vert_main += "  " + it->first + "_cubeCoords = vertex.xyz;\n";
    }

    ///1.6 send 3D coordinates for environment cube map
    if(it->second.GetType() == CUBEMAP_ENV)
    {
       vert_vars += "out vec3 " + g_pref + "r_normal," + g_pref + "r_eyeVec;\n";
       vert_main += 
         "\n  //Camera rotation can be found in transposed modelview matrix\n"
         "  mat3 tcamerarot;\n"
         "  tcamerarot[0][0] = gl_ModelViewMatrixTranspose[0][0];\n"
         "	tcamerarot[0][1] = gl_ModelViewMatrixTranspose[0][1];\n"
         "	tcamerarot[0][2] = gl_ModelViewMatrixTranspose[0][2];\n"
         "	tcamerarot[1][0] = gl_ModelViewMatrixTranspose[1][0];\n"
         "	tcamerarot[1][1] = gl_ModelViewMatrixTranspose[1][1];\n"
         "	tcamerarot[1][2] = gl_ModelViewMatrixTranspose[1][2];\n"
         "	tcamerarot[2][0] = gl_ModelViewMatrixTranspose[2][0];\n"
         "	tcamerarot[2][1] = gl_ModelViewMatrixTranspose[2][1];\n"
         "	tcamerarot[2][2] = gl_ModelViewMatrixTranspose[2][2];\n\n"
         "\n  //Coordinates for cubemap reflexion\n"
         "  " + g_pref + "r_normal = tcamerarot * (gl_NormalMatrix * dNormal);\n"
         "  " + g_pref + "r_eyeVec = tcamerarot * (gl_ModelViewMatrix * vertex).xyz;\n\n";
    }
  }

  //*********************************
  ///1.5 Finalize, vertex transform

  if(use_geom)
		vert_main +=
    "  gl_Position = gl_ModelViewMatrix * vertex;      //vertex transform will be done in geometry shader\n"
    "}\n"
    "\n";
  else
		vert_main +=
    "  gl_Position = gl_ModelViewProjectionMatrix * vertex;      //vertex transform\n"
    "}\n"
    "\n";

  string vertex_shader = "#version 150 compatibility\n";
  vertex_shader += vert_vars + vert_func + vert_main;



  /////////////////////////////////////////////////////////////////////////////
  ///2 CREATION OF GEOMETRY SHADER (OPTIONAL)
  /////////////////////////////////////////////////////////////////////////////

  string geom_vars, geom_func, geom_main;
  //if we render to cube map through geometry shader, add this shader to every material
  if(use_geom)
  {
    ///2.1 Geometry shader variables
    geom_vars = "//GLSL geometry shader generated by gluxEngine\n"
                "#extension GL_EXT_geometry_shader4: enable\n"
                "#extension GL_EXT_gpu_shader4: enable\n"
				"layout(triangles) in;\n"
				"layout(max_vertices=32) out;\n\n";

    geom_vars += "//Material: \"" + name + "\"\n\n";
    geom_func = LoadFunc("matrices");

    //geometry shader main function - sends primitive to each side of cube face defined 
    //in layered framebuffer
    geom_main += LoadFunc("gref_main"); 

    //*****************************
    ///2.2 passing varyings to the fragment shader

    tmp += "Env";
    ///2.2.1 if PHONG, light calculation will be done in fragment shader, so send necessary variables to it (also required for env mapping)
    if(lightModel == PHONG || textures.find(tmp) != textures.end() )
    {
      geom_vars += "in vec3 v_normal[],  v_eyeVec[];\n"  //input variables
                   "out vec3 normal,  eyeVec;\n";        //output variables
      //write normals and eyevec in shader
      geom_main +=
        "      normal = v_normal[i];\n"
        "      eyeVec = v_eyeVec[i];\n";
    }

    //2.2.2 if GOURAUD, send interpolated color to fragment shader
    if(lightModel == GOURAUD)
    {
      geom_vars += "in vec4 v_v_color[];\nout vec4 v_color;";
      geom_main += "      v_color = v_v_color[i];\n";
    }

    //******************************
    //iterate through textures
    for(it = textures.begin(); it != textures.end(); it++)
    {
      ///2.4 send shadow matrices for frag. shader (projected shadow and texture matrix)
      if(it->second.GetType() == SHADOW)
      {
        geom_vars += "in vec4 v_" + it->first + "_projShadow[];\n" +
                     "out vec4 " + it->first + "_projShadow;\n";
        geom_main += "      " + it->first + "_projShadow = v_" + it->first + "_projShadow[i];\n";
      }

      ///2.5 send 3D coordinates for cube map
      if(it->second.GetType() == CUBEMAP)
      {
         geom_vars += "in vec3 v_" + it->first + "_cubeCoords[];\n" +
                      "out vec3 " + it->first + "_cubeCoords;\n";
         geom_main += "      " + it->first + "_cubeCoords = v_" + it->first + "_cubeCoords[i];\n";
      }

      ///2.6 send 3D coordinates for environment cube map
      if(it->second.GetType() == CUBEMAP_ENV)
      {
         geom_vars += "in vec3 v_r_normal[], v_r_eyeVec[];\n"
                      "out vec3 r_normal, r_eyeVec;\n";
         geom_main += 
           "      r_normal = v_r_normal[i];\n"
           "      r_eyeVec = v_r_eyeVec[i];\n";
      }
    }
    //finalize
    geom_main += 
     "      EmitVertex();\n"
     "    }\n"
     "    EndPrimitive();\n"
     "  }\n"
     "}\n";
  }
  string geometry_shader = "#version 150 compatibility\n";
  geometry_shader += geom_vars + geom_func + geom_main;





  ////////////////////////////////////////////////////////////////////////////
  ///3 CREATION OF FRAGMENT SHADER
  ////////////////////////////////////////////////////////////////////////////

  ///3.1 fragment shader variables
  string frag_vars = "//GLSL fragment shader generated by gluxEngine\n\n";
  frag_vars += "//Material: \"" + name + "\"\n\n"
							 "//texture coordinate\n";
  frag_vars +=	"in vec2 fragTexCoord;\n";
  //fragment shader output
  if(useMRT)
	frag_vars += "out vec4 out_FragData[2];\n\n";
  else
	frag_vars += "out vec4 out_FragColor;\n\n";

  string frag_func;                                                         //fragment shader functions
  string frag_main = "\nvoid main()\n{\n";                                  //fragment shader main function


  //**********************************
  ///3.2 Texture pre-setup: create texture samplers for every texture (shadow maps use shadow samplers), texture tiles and coordinates
  frag_main += "  vec2 texCoord = fragTexCoord;\n";
  
  //for parallax/depth map, we modify texture coordinate
  for(it = textures.begin(); it != textures.end(); it++)
  {    
    if(it->second.GetType() == PARALLAX)
    {
      string offset;
      offset = num2str(it->second.intensity);  //parallax offset value
      frag_main +=
            "  //simple parallax mapping, modify texture coordinates\n"
            "  float height = texture2D(" + name + "ParallaxA,  texCoord).r;\n"
            "  float offset = " + offset + " * (2.0 * height - 1.0);\n"
            "  texCoord = texCoord + eyeVec.xy * offset;\n\n";
      break;
    }
 }
        
  for(it = textures.begin(); it != textures.end(); it++)
  {
    if(!it->second.Empty() && it->second.GetType() != SHADOW && it->second.GetType() != DISPLACE )
    {
			frag_main +=
					"\n  vec4 " + it->first + "_texture;\n";

			//cube map, we add different sampler + 3D coordinates
			if(it->second.GetType() == CUBEMAP) 
					frag_vars += "uniform samplerCube "+ it->first + ";\n"
											 "in vec3 " + it->first + "_cubeCoords;\n\n";
			//environment cube map, we need cube sampler and normal + view vector
			else if(it->second.GetType() == CUBEMAP_ENV) 
					frag_vars += "uniform samplerCube "+ it->first + ";\n"
											 "in vec3 r_normal, r_eyeVec;\n\n";
			//regular 2D texture sampler
			else 
			{
				//texture intensity
				frag_vars += "uniform float " + it->first + "_intensity;\n";
				//texture tiles (if set)
				if(it->second.HasTiles())
				{
					frag_vars += "uniform float " + it->first + "_tileX," + it->first + "_tileY;\n";
					frag_main +=
						"  //" + it->first + ", texture tiles\n"
						"  vec2 " + it->first + "_texcoord = texCoord * vec2(" + it->first + "_tileX, " + it->first + "_tileY) ;\n";
				}
				//no tiles
				else
					frag_main += "  vec2 " + it->first + "_texcoord = texCoord;\n";

				frag_vars += "uniform sampler2D "+ it->first + ";\n";
			}
    }
  }


  //**********************************
  ///3.3.1 if light model is set to PHONG, calculate lighting by using varying variables sent from vertex shader (normal, eye vector...)
  if(lightModel == PHONG)    //PHONG, per-pixel
  {  
    frag_vars +=  "in vec3 normal, eyeVec;\n"
				  "//material properties\n"
				  "in vec4 diff, spec;\n"
				  "in float shin;\n";
    frag_func += LoadFunc("light");

    ///3.3.1.1 if is present bump texture, modify normal (bump mapping)
    tmp = name; tmp += "BumpA";
    if( textures.find(tmp) != textures.end() )
    {
      frag_main +=
        "  //bump-mapping\n"
        "  vec3 bump_normal = normal + normalize(texture2D(" + name + "BumpA, " + tmp + "_texcoord).xyz*2.0 - " + name + "BumpA_intensity);\n"
        "  vec4 color = LightModel(bump_normal,eyeVec);\n";   
    }
    else
      frag_main +="  vec4 color = LightModel(normal,eyeVec);\n";
  }
  ///3.3.2 if GOURAUD model is set, only get color value from vertex shader
  else if(lightModel == GOURAUD) 
  {
    frag_vars += "in vec4 v_color;\n"
                 "in vec3 normal;\n";
    frag_main += "  vec4 color = v_color;\n";
  }
  ///3.3.3 else constant shading - only material diffuse is in computation
  else
  {
    frag_vars += "in vec3 normal;\n\n"
			"//material settings (only diffuse)\n"
			"struct Material{\n"
			"  vec3 diffuse;\n};\n"
			"uniform Material material;\n";
    frag_main += "  vec4 color = vec4(material.diffuse,0.0);\n";
  }


  //***************************************************
  ///3.4 texture application according to texture type
	string alpha_test = "";
  for(it = textures.begin(); it != textures.end(); it++)
  {
    if(!it->second.Empty())
    {
			//skip displacement map
			if(it->second.GetType() == DISPLACE)
				continue;
      ///3.4.1 shadow maps - set shadow samplers and call function to project and create soft shadows
      else if(it->second.GetType() == SHADOW)
      {
        frag_vars += "in vec4 " + it->first + "_projShadow;\n";
	    frag_vars +=
          "uniform float " + it->first + "_intensity;\n"
          "uniform sampler2DShadow " + it->first + ";\n";

        //insert shadow function (only once)
        if(it->first.find("ShadowA") != string::npos)
          frag_func += LoadFunc("shadow");

        frag_main += "\n  //Shadow map projection\n"
                     "  color *= PCFShadow(" + it->first + "," + it->first + "_projShadow, " + it->first + "_intensity);\n";
      }
			//other texture types
      else
      {
		//should we use alpha testing?
		if(it->second.GetType() == ALPHA)
			alpha_test = "  //alpha test\n  if( all(lessThan(" + it->first + "_texture.rgb, vec3(0.25)))) discard;\n";
        ///3.4.2 environment maps - add normal and eye vector variables(if per-pixel)
        if(it->second.GetType() == ENV)
        {
          //when we have more env maps, insert env function only once
          if(it->first.find("EnvB") == string::npos)
          {

            if(lightModel != PHONG)
              frag_vars += "in vec3 eyeVec;\n";
            //add environment map computation
            frag_func += LoadFunc("env");
          }
        }

        //reflect fragment from environment cubemap
        if(it->second.GetType() == CUBEMAP_ENV)
          frag_main += "  vec3 reflVec = reflect(normalize(r_eyeVec), normalize(r_normal)); //reflection vector for cubemap\n";

        //compute fragment color from texel(not for bump/parallax map)
        if(it->second.GetType() != BUMP && it->second.GetType() != PARALLAX)
          frag_main += computeTexel(it->second,it->first);

      }
    }
  }


  //***********************************************
  ///3.5 Finalize (with transparency or alpha test)
  frag_main +=
    "\n  //FINAL fragment color\n";

	//do we use MRT?
	if(useMRT)
	{
		//is material transparent?
		if(transparency > 0.0)
		{
			string trans;
			trans = num2str(transparency);
			frag_main += "  out_FragData[0] = vec4(vec3(color.rgb),  " + trans + ");\n";
								   "  out_FragData[1].rgb = normal;    //normal;\n"
                   "  out_FragData[1].a = pow(gl_FragCoord.z,128.0);   //depth;\n";
		}
		//alpha test fragment
		else
		{
			//alpha test fragment
			frag_main += alpha_test + "  out_FragData[0].rgb = color.rgb; //color;\n"
                   "  out_FragData[1].rgb = normal;    //normal;\n"
                   "  out_FragData[1].a = pow(gl_FragCoord.z,128.0);   //depth;\n";
		}
	}
	else
	{
		//is material transparent?
		if(transparency > 0.0)
		{
			string trans;
			trans = num2str(transparency);
			frag_main += "  out_FragColor = vec4(vec3(color.rgb),  ";
			frag_main += trans;
			frag_main += ");\n";
		}
		else
		{
			//alpha test fragment
			frag_main += alpha_test + "  out_FragColor = color;\n";
		}
	}
/*
  //is material transparent?
  if(transparency > 0.0)
  {
    string trans;
    trans = num2str(transparency);
    frag_main += "  out_FragColor = vec4(vec3(color.rgb),  ";
    frag_main += trans;
    frag_main += ");\n";
  }
  //for MRT, output also normal colors
  else if(useMRT)
    frag_main += "  out_FragData[0].rgb = color.rgb; //color;\n"
                 "  out_FragData[1].rgb = normal;    //normal;\n"
                 "  out_FragData[1].a = pow(gl_FragCoord.z,128.0);   //depth;\n";
  else
	{
		//alpha test fragment
		frag_main += alpha_test + "  out_FragColor = color;\n";
	}
*/
  frag_main +=
    "}\n"
    "\n";
  string frag_shader = "#version 150 compatibility\n";
  frag_shader += frag_vars + frag_func + frag_main;

  //////////////////////////////////////////////////////////////////////////////


  //save shaders to debug file
  string file; file = "shader_out/" + name + ".vert";
  ofstream fout1(file.c_str());
  fout1<<vertex_shader;
  file = "shader_out/" + name + ".frag";
  ofstream fout2(file.c_str());
  fout2<<frag_shader;

  ///4 Create, compile and link shaders
  v_shader = glCreateShader(GL_VERTEX_SHADER);
  f_shader = glCreateShader(GL_FRAGMENT_SHADER);

  //copy shader data
  const char *ff = frag_shader.c_str();
  const char *vv = vertex_shader.c_str();

  //set shader source
  glShaderSource(v_shader, 1, &vv,NULL);
  glShaderSource(f_shader, 1, &ff,NULL);

  //compile and detect shader errors
  char log[BUFFER]; int len;
  glCompileShader(v_shader);
  glCompileShader(f_shader);

  //create and link shader program
  GLuint tmp_shader = glCreateProgram();
  glAttachShader(tmp_shader,f_shader);
  glAttachShader(tmp_shader,v_shader);

  //add geometry shader if present
  if(use_geom)
  {
    file = "shader_out/" + name + ".geom";
    ofstream fout3(file.c_str());
    fout3<<geometry_shader;    

    g_shader = glCreateShader(GL_GEOMETRY_SHADER);
    const char *gg = geometry_shader.c_str();
    glShaderSource(g_shader, 1, &gg,NULL);
    glCompileShader(g_shader);
    glAttachShader(tmp_shader,g_shader);
  }
  glLinkProgram(tmp_shader);
  glUseProgram(tmp_shader);


  //shader creation status
  glGetShaderInfoLog(v_shader, BUFFER, &len, log);
  if(strstr(log, "succes") == NULL && len > 0) 
		cout<<endl<<name<<":"<<log;    //print error if any
  if(use_geom)
  {
    glGetShaderInfoLog(g_shader, BUFFER, &len, log);
    if(strstr(log, "succes") == NULL && len > 0) 
			cout<<endl<<name<<":"<<log;    
  }
  glGetShaderInfoLog(f_shader, BUFFER, &len, log);
  if(strstr(log, "succes") == NULL && len > 0) 
		cout<<endl<<name<<":"<<log;    //print error if any


  //*************************************
  ///4 Get uniform variables for textures (using Texture::GetUniforms() )
	int i=0;
  for(it = textures.begin(); it != textures.end(); it++)
  {
    if(!it->second.Empty())
    {
      it->second.GetUniforms(tmp_shader);
	  it->second.ActivateTexture(i,true);
	  i++;
    }
  }

#ifdef GL3
	//Bind generic vertex attributes: vertices, normals and coordinates
	glBindAttribLocation(tmp_shader, 0, "in_Vertex");
  glBindAttribLocation(tmp_shader, 1, "in_Normal");
	glBindAttribLocation(tmp_shader, 3, "in_Coord");
#endif

  if(use_geom) 
		gshader_ref = tmp_shader;   //geometry shader for reflection
  else 
		shader = tmp_shader;                       //normal shader

  ///set materials parameters
	SetUniforms("material.ambient",ambColor, 3);
	SetUniforms("material.diffuse",diffColor, 3);
	SetUniforms("material.specular",specColor, 3);
	SetUniform("material.shininess",shininess);

  baked = true;

  return true;
}
