/**
****************************************************************************************************
****************************************************************************************************
@file: render_target.cpp
@brief handles rendering to texture/cube maps for dynamic reflections/refractions
****************************************************************************************************
***************************************************************************************************/
#include "scene.h"


/**
****************************************************************************************************
@brief Create render target using framebuffer and attached texture
@param resX X-size of framebuffer texture
@param resY Y-size of framebuffer texture
@param tex_format OpenGL texture format (can be GL_RGBA, GL_RGBA16F or GL_RGBA32F)
@param tex_type internal data format (can be GL_UNSIGNED_BYTE or GL_FLOAT)
***************************************************************************************************/
void TScene::CreateRenderTarget(int resX, int resY, GLint tex_format, GLenum tex_type )
{
  //if no resolution given, use current scene resolution
  if(resX == -1 || resY == -1)
  {
    resX = resx;
    resY = resy;
  }

	//save render target size
	RT_resX = resX;
	RT_resY = resY;

  //create texture - for original image
  glGenTextures(1, &render_texture);
  glBindTexture(GL_TEXTURE_2D, render_texture);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, tex_format, resX, resY, 0, GL_RGBA, tex_type, NULL);

  //create texture - for bloom effect
  glGenTextures(1, &bloom_texture);
  glBindTexture(GL_TEXTURE_2D, bloom_texture);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, tex_format, resX, resY, 0, GL_RGBA, tex_type, NULL);

  //create texture - for blur
  glGenTextures(1, &blur_texture);
  glBindTexture(GL_TEXTURE_2D, blur_texture);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, tex_format, resX, resY, 0, GL_RGBA, tex_type, NULL);

  //create texture - for store normal values
  glGenTextures(1, &normal_texture);
  glBindTexture(GL_TEXTURE_2D, normal_texture);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, tex_format, resX, resY, 0, GL_RGBA, tex_type, NULL);

  glGenRenderbuffers(1, &r_buffer_depth);
  glBindRenderbuffer(GL_RENDERBUFFER, r_buffer_depth);
  glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT,resX, resY);

  //if MSAA enabled, create also multisampled framebuffer (only when platform supports it)
  if(msamples > 1 && GLEW_EXT_framebuffer_multisample)
  {
    //Create multisampled color renderbuffer
    glGenRenderbuffers(1, &msaa_r_buffer_color);
    glBindRenderbuffer(GL_RENDERBUFFER, msaa_r_buffer_color);
    glRenderbufferStorageMultisample(GL_RENDERBUFFER, msamples, tex_format, resX, resY);

    //Create multisampled depth renderbuffer
    glGenRenderbuffers(1, &msaa_r_buffer_depth);
    glBindRenderbuffer(GL_RENDERBUFFER, msaa_r_buffer_depth);
    glRenderbufferStorageMultisample(GL_RENDERBUFFER, msamples, GL_DEPTH_COMPONENT,resX, resY);

    //Attach to multisampled framebuffer
    glGenFramebuffers(1, &msaa_f_buffer);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, msaa_f_buffer);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_RENDERBUFFER_EXT, msaa_r_buffer_color);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, msaa_r_buffer_depth);

    //check FBO creation
    if( glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
    {
	    if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_UNDEFINED) 
		    cout<<"FBO ERROR:GL_FRAMEBUFFER_UNDEFINED\n";
	    if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT) 
		    cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT\n";
	    if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT) 
		    cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT\n";
	    if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT) 
		    cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER\n";
	    if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT) 
		    cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER\n";
	    if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE_EXT) 
		    cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE\n";
	    if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS_EXT) 
		    cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS\n";
	    if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_UNSUPPORTED_EXT) 
		    cout<<"FBO ERROR:GL_FRAMEBUFFER_UNSUPPORTED\n";
      cout<<"WARNING: Multisampled FBO creation failed, falling back to regular FBO.\n";
      msamples = 1;
    }
  }

  //Finally, create framebuffer
  glGenFramebuffers(1, &f_buffer);
  glBindFramebufferEXT(GL_FRAMEBUFFER, f_buffer);

  //attach texture to the frame buffer
  glFramebufferTexture2D(GL_FRAMEBUFFER,GL_COLOR_ATTACHMENT0,GL_TEXTURE_2D, render_texture, 0);
  //attach also normal texture (only if SSAO enabled)
  if(useSSAO)
    glFramebufferTexture2D(GL_FRAMEBUFFER,GL_COLOR_ATTACHMENT1,GL_TEXTURE_2D, normal_texture, 0);
  //attach render buffers
	//glFramebufferRenderbuffer(GL_FRAMEBUFFER,GL_COLOR_ATTACHMENT0,GL_RENDERBUFFER, r_buffer_color);
  glFramebufferRenderbuffer(GL_FRAMEBUFFER,GL_DEPTH_ATTACHMENT,GL_RENDERBUFFER, r_buffer_depth);

  //check FBO creation
  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
  {
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_UNDEFINED) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_UNDEFINED\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_UNSUPPORTED_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_UNSUPPORTED\n";
    ShowMessage("ERROR: FBO creation failed!",false);
    useHDR = useSSAO = false;
    throw ERR;
  }



  // Go back to regular frame buffer rendering
  glBindFramebufferEXT(GL_FRAMEBUFFER, 0);

  //add created textures into cache
  tex_cache["render_texture"] = render_texture;
  tex_cache["bloom_texture"] = bloom_texture;
  tex_cache["normal_texture"] = normal_texture;
  tex_cache["blur_texture"] = blur_texture;

	//update load list
	UpdateLoadList(17);

  //create screen aligned quad
  AddObject("screen_quad",SCREEN_QUAD);

  //HDR passes - render to texture, bloom effect and tonemapping
  AddMaterial("mat_bloom_hdr",white,white,white,0.0,0.0,0.0,NONE);  //HDR only
  AddMaterial("mat_bloom_ssao",white,white,white,0.0,0.0,0.0,NONE);  //SSAO only
  AddMaterial("mat_bloom_hdr_ssao",white,white,white,0.0,0.0,0.0,NONE);  //HDR + SSAO

  //HDR only
  AddTexture("mat_bloom_hdr","render_texture",RENDER_TEXTURE);
  CustomShader("mat_bloom_hdr","data/shaders/quad.vert",NULL,"data/shaders/bloom_ssao.frag",NULL,"#define BLOOM\n");

  //SSAO only
  AddTexture("mat_bloom_ssao","normal_texture",NORMAL_TEXTURE);
  AddTexture("mat_bloom_ssao","data/tex/random.tga");
  CustomShader("mat_bloom_ssao","data/shaders/quad.vert",NULL,"data/shaders/bloom_ssao.frag",NULL,"#define SSAO\n");

  //HDR + SSAO
  AddTexture("mat_bloom_hdr_ssao","render_texture",RENDER_TEXTURE);
  AddTexture("mat_bloom_hdr_ssao","normal_texture",NORMAL_TEXTURE);
  AddTexture("mat_bloom_hdr_ssao","data/tex/random.tga");
  CustomShader("mat_bloom_hdr_ssao","data/shaders/quad.vert",NULL,"data/shaders/bloom_ssao.frag",NULL,"#define BLOOM\n#define SSAO\n");

  //tonemapping
  AddMaterial("mat_tonemap",white,white,white,0.0,0.0,0.0,NONE);
  AddTexture("mat_tonemap","render_texture",RENDER_TEXTURE);
  AddTexture("mat_tonemap","blur_texture",BLUR_TEXTURE);
  CustomShader("mat_tonemap","data/shaders/quad.vert",NULL,"data/shaders/tonemap.frag");

  //blur
  AddMaterial("mat_blur_horiz",white,white,white,0.0,0.0,0.0,NONE);
  AddTexture("mat_blur_horiz","bloom_texture",BLOOM_TEXTURE);
	CustomShader("mat_blur_horiz","data/shaders/quad.vert",NULL,"data/shaders/blur.frag",NULL,"#define HORIZONTAL\n");
  AddMaterial("mat_blur_vert",white,white,white,0.0,0.0,0.0,NONE);
  AddTexture("mat_blur_vert","bloom_texture",BLOOM_TEXTURE);
  CustomShader("mat_blur_vert","data/shaders/quad.vert",NULL,"data/shaders/blur.frag","#define VERTICAL\n");
}

/**
****************************************************************************************************
@brief Resize render target textures
@param resX new X-size of framebuffer textures
@param resY new Y-size of framebuffer texture
@param tex_format OpenGL texture format (can be GL_RGBA, GL_RGBA16F or GL_RGBA32F)
@param tex_type internal data format (can be GL_UNSIGNED_BYTE or GL_FLOAT)
***************************************************************************************************/
void TScene::ResizeRenderTarget(int resX, int resY, GLint tex_format, GLenum tex_type )
{
  //FBO must be attached, otherwise return
  if(f_buffer == NULL) return;

  //resize texture - for original image
  glBindTexture(GL_TEXTURE_2D, render_texture);
  glTexImage2D(GL_TEXTURE_2D, 0, tex_format, resX, resY, 0, GL_RGBA, tex_type, NULL);

  //resize texture - for bloom effect
  glBindTexture(GL_TEXTURE_2D, bloom_texture);
  glTexImage2D(GL_TEXTURE_2D, 0, tex_format, resx, resy, 0, GL_RGBA, tex_type, NULL);

  //resize texture - for blur
  glBindTexture(GL_TEXTURE_2D, blur_texture);
  glTexImage2D(GL_TEXTURE_2D, 0, tex_format, resx, resy, 0, GL_RGBA, tex_type, NULL);

  //resize texture - for store normal values
  glBindTexture(GL_TEXTURE_2D, normal_texture);
  glTexImage2D(GL_TEXTURE_2D, 0, tex_format, resx, resy, 0, GL_RGBA, tex_type, NULL);

  //resize renderbuffer storage
  glBindRenderbuffer(GL_RENDERBUFFER, r_buffer_depth);
  glRenderbufferStorage(GL_RENDERBUFFER,GL_DEPTH_COMPONENT32F,resx, resy);

  glBindFramebufferEXT(GL_FRAMEBUFFER, r_buffer_depth);
  glFramebufferRenderbuffer(GL_FRAMEBUFFER,GL_DEPTH_ATTACHMENT,GL_RENDERBUFFER, r_buffer_depth);
  glBindFramebufferEXT(GL_FRAMEBUFFER, 0);
}

/**
****************************************************************************************************
@brief Prepare cube map texture for rendering into
@param obj pointer to object with reflection
@return succesful creation
***************************************************************************************************/
bool TScene::CreateRenderCubeMap(TObject *obj)
{
  cout<<"Creating dynamic cube map for \""<<obj->GetName()<<"\"...\n";

  //delete possible old textures and FBO
	//if(obj->HasFramebuffer())
	{
		glDeleteFramebuffers(1, &obj->cf_buffer);
		glDeleteRenderbuffers(1, &obj->cr_buffer);
		glDeleteTextures(1, &obj->render_cubemap);
	}

  //create texture
  glGenTextures(1, &obj->render_cubemap);
  glBindTexture(GL_TEXTURE_CUBE_MAP, obj->render_cubemap);

  //fill 6 sides of texture with null data
  for(int i=0; i<6; i++)
    glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_RGBA, obj->refl_size, obj->refl_size, 0, GL_RGB, GL_UNSIGNED_BYTE, NULL);

  //texture parameters
  glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE );
  glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE );
  glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
  glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
  //glBindTexture(GL_TEXTURE_CUBE_MAP, 0);

  //Create renderbuffer for depth
  glGenRenderbuffers(1, &obj->cr_buffer);
  glBindRenderbuffer(GL_RENDERBUFFER, obj->cr_buffer);
  glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_COMPONENT, obj->refl_size, obj->refl_size);

  //Create FBO
  glGenFramebuffers(1, &obj->cf_buffer);
  glBindFramebufferEXT(GL_FRAMEBUFFER, obj->cf_buffer);

  //if geometry shader is used, attach only one texture side as layer(layers will be written in geometry shader)
  if(use_gshader_ref)
  {
    //create addition depth cubetexture
    glGenTextures(1, &obj->render_cubemap_depth);
    glBindTexture(GL_TEXTURE_CUBE_MAP, obj->render_cubemap_depth);

    //fill 6 sides of texture with null data
    for(int i=0; i<6; i++)
      glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_DEPTH_COMPONENT, obj->refl_size, obj->refl_size, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, NULL);

    //texture parameters
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    glTexParameteri( GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    //glBindTexture(GL_TEXTURE_CUBE_MAP, 0);

    glFramebufferTextureARB(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, obj->render_cubemap, 0);
    glFramebufferTextureARB(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, obj->render_cubemap_depth, 0);
  }
  //else attach all six cube faces as FBO color attachments
  else
  {
    for(int i=0; i<6; i++)
      glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0+i, GL_TEXTURE_CUBE_MAP_POSITIVE_X+i, obj->render_cubemap, 0);
    //attach depth renderbuffer to FBO
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_RENDERBUFFER, obj->cr_buffer);
  }


  //check FBO creation
  if(glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
  {
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_UNDEFINED) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_UNDEFINED\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_ATTACHMENT\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_MISSING_ATTACHMENT\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_DRAW_BUFFER\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_READ_BUFFER\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_MULTISAMPLE\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_INCOMPLETE_LAYER_TARGETS\n";
	  if( glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT) ==  GL_FRAMEBUFFER_UNSUPPORTED_EXT) 
		  cout<<"FBO ERROR:GL_FRAMEBUFFER_UNSUPPORTED\n";
    ShowMessage("ERROR: Framebuffer object creation failed!",false);
    return false;
  }
  glBindFramebufferEXT(GL_FRAMEBUFFER, 0);

  string texname = obj->GetName() + "Cubemap";
  tex_cache[texname] = obj->render_cubemap;

  //add cubemap to material
  cout<<"Adding cube texture to material \""<<obj->GetMatName()<<"\"\n";
  materials[obj->GetMatName()].AddTexture("dynamic_cube",CUBEMAP_ENV,ADD,1.0,1.0,1.0,true,false,obj->render_cubemap);
  return true;
}


/**
****************************************************************************************************
@brief Render scene to cubemap using six framebuffers
@param obj object with dynamic cubemap
***************************************************************************************************/
void TScene::RenderToCubeMap(TObject *obj)
{
  //push old viewport
  glPushAttrib(GL_VIEWPORT_BIT);
  glViewport(0, 0, obj->refl_size, obj->refl_size);

  //bind framebuffer to render into
  glBindFramebufferEXT(GL_FRAMEBUFFER, obj->cf_buffer);

  //set new projection matrix
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();

  TVector pos = obj->GetPosition();
  string obj_name = obj->GetName();

  //camera fov = 90 degrees, aspect ratio = 1.0 - render one side of cube
  gluPerspective (90.0f, 1.0, 0.5, 500.0);

  //in geometry shader we draw only in one pass - shader takes care of the rest
  if(use_gshader_ref)
  {
    //turn on geometry shading
    for(im = materials.begin(); im != materials.end(); im++)
      im->second.UseGShaderCubeMapRender(true);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    //translate camera to the center of the reflective object
    glTranslatef(-pos.x,-pos.y,-pos.z);
    glGetFloatv(GL_MODELVIEW_MATRIX, tmpMatrix);

    //glDrawBuffer(GL_COLOR_ATTACHMENT2);
    //glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    //draw whole scene except our reflecting object
    for(il = lights.begin(); il != lights.end(); il++)
       il->KeepPos();

    map<string,TObject>::iterator ii;
    for(ii = objects.begin(); ii != objects.end(); ii++)
			if(ii->second.GetSceneID() == sceneID)
				if(ii->first != obj_name)
					ii->second.Draw(viewMatrix);

    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();

    //turn off geometry shading
    for(im = materials.begin(); im != materials.end(); im++)
      im->second.UseGShaderCubeMapRender(false);
  }

  //otherwise we must "manually" render six views to each face of cube
  else
  {
    for(int i=0; i<6; i++)
    {
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      glLoadIdentity();

      //rotate view to every side of cube
      switch(i)
      {
        case 0: glRotatef( 90.0,0.0,1.0,0.0); glRotatef( 180.0,1.0,0.0,0.0); break;//positive X
        case 1: glRotatef(-90.0,0.0,1.0,0.0); glRotatef( 180.0,1.0,0.0,0.0); break;//negative X
        case 2: glRotatef(-90.0,1.0,0.0,0.0);  break;//positive Y
        case 3: glRotatef( 90.0,1.0,0.0,0.0);  break;//negative Y
        case 4: glRotatef(180.0,0.0,1.0,0.0); glRotatef( 180.0,0.0,0.0,1.0); break;//positive Z
        case 5: glRotatef( 180.0,0.0,0.0,1.0); break; //negative Z
      }
      //translate camera to the center of the reflective object
      glTranslatef(-pos.x,-pos.y,-pos.z);
	    glGetFloatv(GL_MODELVIEW_MATRIX, tmpMatrix);

      glDrawBuffer(GL_COLOR_ATTACHMENT0+i);
      glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

      //draw whole scene except our reflecting object
      for(il = lights.begin(); il != lights.end(); il++)
        il->KeepPos();

      map<string,TObject>::iterator ii;
      for(ii = objects.begin(); ii != objects.end(); ii++)
				if(ii->second.GetSceneID() == sceneID)
					if(ii->first != obj_name)
						ii->second.Draw(tmpMatrix);

      glMatrixMode(GL_MODELVIEW);
      glPopMatrix();
    }
  }

  glBindFramebufferEXT(GL_FRAMEBUFFER, 0);
  glDrawBuffer(GL_BACK);
  //restore projection matrix
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  //restore viewport
  glPopAttrib();
}
