/**
****************************************************************************************************
****************************************************************************************************
@file: scene.cpp
@brief initialization and drawing the scene - definitions
****************************************************************************************************
***************************************************************************************************/
#include "scene.h"

/**
****************************************************************************************************
@brief Initialize scene by clearing lists
****************************************************************************************************/
TScene::TScene()
{
  lights.clear();
  useHDR = useSSAO = false;
  use_gshader_ref = false;
	f_buffer = r_buffer_color = r_buffer_depth = msamples = 0;
	load_list = load_actual = 0;
	testname = "...";
	sceneID = 0;
}

/**
****************************************************************************************************
@brief Destructor. Free dynamic data
****************************************************************************************************/
TScene::~TScene()
{
	Destroy(true);
	//delete font
	glDeleteLists(font2D,256);
}

/**
****************************************************************************************************
@brief Scene first initialization. In this function it's being checked if accelerator has all necessary
extensions to run program. Then are set basic scene variables like screen resolution, near and far
clipping planes and perspective correction.
@param _resx screen width in pixels
@param _resy screen height in pixels
@param _near near clipping plane (must be greater than 0.0)
@param _far far clipping plane (must be greater than near)
@param _fovy viewing angle (in degrees)
@param _msamples number of samples used in multisampling
@param cust_cam it's possible to use a custom camera instead of scene camera
@param load_font shall we load font?
@return success/fail of initialization
****************************************************************************************************/
bool TScene::PreInit(GLint _resx, GLint _resy, GLfloat _near, GLfloat _far, GLfloat _fovy, int _msamples, bool cust_cam, bool load_font)
{
  resx = _resx;
  resy = _resy;
  near_p = _near;
  far_p = _far;
  fovy = _fovy;
  custom_cam = cust_cam;
  msamples = _msamples;

  glPixelStorei(GL_UNPACK_ALIGNMENT, 4);      // 4-byte pixel alignment
  glClearColor(0.0, 0.0, 0.0, 0.5); //clear color and depth
  glClearDepth(1.0);
  glEnable(GL_DEPTH_TEST);          //enable depth buffer
  glDepthFunc(GL_LEQUAL);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);    //perspective correction
  glEnable(GL_TEXTURE_CUBE_MAP_SEAMLESS);


  //create initial scene size
  Resize(resx,resy);

  //initialize font
	if(load_font)
		BuildFont();

  ///at the end of init add basic material to list
  //AddMaterial("basic");
  return true;
}


/**
****************************************************************************************************
@brief Scene second initialization. The purpose of this initialization is to assign default material
to objects without material and create shadow maps(TScene::CreateShadowMap()) for lights with shadows
enabled, and to add this shadow map(TMaterial::AddShadowMap()) to all materials
@return success/fail of initialization
****************************************************************************************************/
bool TScene::PostInit()
{
  //do we have shadows from lights?
  for(il = lights.begin(); il != lights.end(); il++)
  {
    if(il->HasShadow()) //if yes, create shadow map for current light
    {
      if(!CreateShadowMap(il))
				return false;

			//add shadow map to all materials (except those who don't receive shadows)
			for(im = materials.begin(); im != materials.end(); im++)
				if(im->second.GetSceneID() == sceneID)
				 im->second.AddShadowMap(tex_cache[il->GetShadowCache()], il->ShadowIntensity() );
    }
  }

  cout<<"Baking materials...\n";
  //assign light count to all materials and then bake them
  for(im = materials.begin(); im != materials.end(); im++)
  {
		if(im->second.GetSceneID() == sceneID)
		{
			//set MRT if we use SSAO effect
			if(useSSAO)
				im->second.UseMRT(true);
			//bake material
			im->second.BakeMaterial();
			im->second.SetLights(lights.size());
		}
  }
	cout<<"Post Init OK\n";

  return true;
}


/**
****************************************************************************************************
@brief Resize screen to new dimensions, re-initialize viewport and projection matrices
@param _resx X-resolution in pixels
@param _resy Y-resolution in pixels
***************************************************************************************************/
void TScene::Resize(GLint _resx, GLint _resy)
{
  resx = _resx;
  resy = _resy;
  glViewport(0,0,resx,resy);        //new viewport settings
  glMatrixMode(GL_PROJECTION);      //projection matrix
  glLoadIdentity();
  //perspective settings
  gluPerspective(fovy,(GLfloat)resx/resy,near_p,far_p);
  glMatrixMode(GL_MODELVIEW);       //new modelview matrix
  glLoadIdentity();
}




/**
****************************************************************************************************
@brief Render scene - camera, lights and objects(run through object list and call render function
for every object)
@param delete_buffer should we delete color and depth buffer?
***************************************************************************************************/
void TScene::Redraw(bool delete_buffer)
{
  //HDR/SSAO renderer - render to texture
  if(useHDR || useSSAO)
  {
		//render target viewport size
		glViewport(0,0,RT_resX,RT_resY);

    //attach framebuffer to render to
    glBindFramebufferEXT(GL_FRAMEBUFFER, f_buffer);
    //attach render texture
    glFramebufferTexture2D(GL_FRAMEBUFFER,GL_COLOR_ATTACHMENT0,GL_TEXTURE_2D, render_texture, 0);

    ///@todo multisample FBO doesn't work with MRT
    if(msamples > 1 && GLEW_EXT_framebuffer_multisample && !useSSAO)
      glBindFramebufferEXT(GL_FRAMEBUFFER, msaa_f_buffer);  //render to multisample FBO

    //multiple render targets - only when using SSAO
    if(useSSAO)
    {
      GLenum mrt[] = { GL_COLOR_ATTACHMENT0, GL_COLOR_ATTACHMENT1 };
      glDrawBuffers(2, mrt);
    }

    //clear screen (if desired)
    if(delete_buffer)
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
  }
  else        //else render to default framebuffer, clear it(if desired)
  {
    if(delete_buffer)
      glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
  }

  if(!custom_cam)
		cam.RenderCamera();   //it's possible to use custom camera settings

  ///position all lights,
  for(il = lights.begin(); il != lights.end(); il++)
  {
    il->KeepPos();
    il->DrawLight(viewMatrix);

    ///if light has a shadow, render scene from light view to texture (TScene::RenderShadowMap())
    if(il->HasShadow())
    {
      //unbind framebuffer if bound to
      if(useHDR || useSSAO)
				glBindFramebufferEXT(GL_FRAMEBUFFER, 0);
			//then render shadow map
      RenderShadowMap(*il);
      if(useHDR || useSSAO)
			{
				//re-bind original FBO (or multisampled)
				if(msamples > 1 && GLEW_EXT_framebuffer_multisample && !useSSAO)
					glBindFramebufferEXT(GL_FRAMEBUFFER, msaa_f_buffer);
				else
					glBindFramebufferEXT(GL_FRAMEBUFFER, f_buffer);
			}
    }
  }

	//send some uniforms to all shaders in scene
  for(im = materials.begin(); im != materials.end(); im++)
	{
		im->second.SetUniforms("camPos",cam.GetPos(),3);	//camera position
		if(lights.size() > 0)
			im->second.SetUniform("lightRadius",lights[0].GetRadius());	//light radius
	}

  ///render all opaque objects (TObject::Draw())
  for(io = objects.begin(); io != objects.end(); io++)
		if(io->second.GetSceneID() == sceneID)
			if(!io->second.IsTransparent())
				io->second.Draw(viewMatrix);

  ///then render all transparent objects (TObject::Draw())
  glBlendFunc(GL_SRC_ALPHA, GL_ONE);
  glEnable(GL_BLEND);
  for(io = objects.begin(); io != objects.end(); io++)
		if(io->second.GetSceneID() == sceneID)
			if(io->second.IsTransparent())
				io->second.Draw(viewMatrix);
	glDisable(GL_BLEND);

  ///render dynamic cubemaps for reflective objects
  for(io = objects.begin(); io != objects.end(); io++)
		if(io->second.GetSceneID() == sceneID)
			//do we have objects with reflection? If yes, generate cubemaps for it
			if(io->second.GetReflect() > 0.0)
				RenderToCubeMap(&io->second);
  //HDR/SSAO renderer
  if(useHDR || useSSAO)
  {
    //if MSAA enabled, copy from multisampled FBO to normal FBO
    ///@todo multisample FBO doesn't work with MRT
    if(msamples > 1 && GLEW_EXT_framebuffer_multisample && !useSSAO)
    {
      glBindFramebufferEXT(GL_READ_FRAMEBUFFER_EXT, msaa_f_buffer);
      glBindFramebufferEXT(GL_DRAW_FRAMEBUFFER_EXT, f_buffer);
      glBlitFramebufferEXT(0, 0, resx, resy, 0, 0, resx, resy, GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT, GL_NEAREST);
    }

    glBindFramebufferEXT(GL_FRAMEBUFFER, 0);
    glClear(GL_DEPTH_BUFFER_BIT);
    glViewport(0,0,RT_resX,RT_resY);

    //scene is now rendered in hdr_texture - extract brightest parts
    glBindFramebufferEXT(GL_FRAMEBUFFER, f_buffer);
    //now render only one target
    GLenum srt[] = { GL_COLOR_ATTACHMENT0 };
    glDrawBuffers(1, srt);

    //attach bloom texture
    glFramebufferTexture2D(GL_FRAMEBUFFER,GL_COLOR_ATTACHMENT0,GL_TEXTURE_2D, bloom_texture, 0);

    //choose material - we can render HDR only, SSAO only or both together
    if( useHDR && !useSSAO)
			SetMaterial("screen_quad","mat_bloom_hdr");
    if(!useHDR &&  useSSAO)
			SetMaterial("screen_quad","mat_bloom_ssao");
    if( useHDR &&  useSSAO)
			SetMaterial("screen_quad","mat_bloom_hdr_ssao");

    //downsample bloom texture by setting new viewport
    glViewport(0,0,RT_resX/2,RT_resY/2);
    //draw brightest parts to bloom texture
    objects["screen_quad"].DrawScreenQuad();

    //horizontal blur
    SetMaterial("screen_quad","mat_blur_horiz");
    objects["screen_quad"].DrawScreenQuad();

		//vertical blur
    glFramebufferTexture2D(GL_FRAMEBUFFER,GL_COLOR_ATTACHMENT0,GL_TEXTURE_2D, blur_texture, 0);
    SetMaterial("screen_quad","mat_blur_vert");
    objects["screen_quad"].DrawScreenQuad();

    glBindFramebufferEXT(GL_FRAMEBUFFER, 0);

    //final draw with bloom and tone mapping
    glViewport(0,0,resx,resy);  //restore original scene viewport

    SetMaterial("screen_quad","mat_tonemap");
    objects["screen_quad"].DrawScreenQuad();
  }
}

/**
****************************************************************************************************
@brief Draw loading screen
***************************************************************************************************/
void TScene::LoadScreen(bool swap)
{
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	//background
	materials["mat_background_quad"].RenderMaterial();
  glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(0.0f,1.0f); glVertex2f(-1.0f, 1.0f);
    glTexCoord2f(1.0f,1.0f); glVertex2f( 1.0f, 1.0f);
    glTexCoord2f(0.0f,0.0f); glVertex2f(-1.0f,-1.0f);
    glTexCoord2f(1.0f,0.0f); glVertex2f( 1.0f,-1.0f);
  glEnd();

	DrawScreenText(testname.c_str(),50.0,50.0,1.5);
	DrawScreenText("Initializing, please wait... ",50.0f,45.0f,1.2f);

	materials["mat_load_polygon"].RenderMaterial();

	//progress bar
	load_actual--;
	float loaded;
	if(load_actual > 0)
		loaded = -0.7f + (float)(load_list - load_actual)/load_list;
	else
		loaded = 0.7f;

  //draw quad
  glBegin(GL_TRIANGLE_STRIP);
    glTexCoord2f(0.0f,1.0f); glVertex2f(-0.7f,-0.2f);
    glTexCoord2f(1.0f,1.0f); glVertex2f( loaded,-0.2f);
    glTexCoord2f(0.0f,0.0f); glVertex2f(-0.7f,-0.3f);
    glTexCoord2f(1.0f,0.0f); glVertex2f( loaded,-0.3f);
  glEnd();

	if(swap)
		SDL_GL_SwapBuffers();
}

/**
****************************************************************************************************
@brief Destroys all dynamic scene data, but keeps scene settings (like perspective, clipping planes)
@param delete_cache should we delete also loaded textures from cache? (can be useful when next scene
uses the same textures)
***************************************************************************************************/
void TScene::Destroy(bool delete_cache)
{
  objects.clear();
  materials.clear();
  lights.clear();
  cam.Reset();

  if(delete_cache)
	{
		tex_cache.clear();
		obj_cache.clear();
	}

  //reset framebuffers
	if(useHDR || useSSAO)
	{
		glDeleteFramebuffers(1, &f_buffer);
		glDeleteRenderbuffers(1, &r_buffer_color);
		glDeleteRenderbuffers(1, &r_buffer_depth);
		glDeleteTextures(1,&render_texture);
		glDeleteTextures(1,&bloom_texture);
		glDeleteTextures(1,&normal_texture);
		glDeleteTextures(1,&blur_texture);
	}
/*
	f_buffer = r_buffer_color = r_buffer_depth = 0;
	useHDR = useSSAO = use_gshader_ref = useHDR = useSSAO = useShadows = false;
*/
  //reset camera
  cam.Reset();
}


/**
****************************************************************************************************
@brief Toggle using of dynamic shader cubemapping
@param use_gs on/off geometry shader for cubemapping
***************************************************************************************************/
void TScene::UseDynReflections(bool use_gs)
{
  use_gshader_ref = use_gs;

  //create cubemaps
	for(io = objects.begin(); io != objects.end(); io++)
	{
		if(io->second.GetReflect() > 0.0)
			if(!CreateRenderCubeMap(&io->second))
				throw ERR;
	}

  //toggle geometry shading in materials
  for(im = materials.begin(); im != materials.end(); im++)
	{
		if(use_gshader_ref)
		{
			im->second.UseGShaderCubeMapRender(true);
			if(!im->second.BakeMaterial(true))						//create additional geometry shader
				throw ERR;
		}
	}
}


/**
****************************************************************************************************
@brief Return camera position in world space
@return camera position
***************************************************************************************************/
/*
TVector TScene::GetWorldCameraPos()
{
  double cposx,cposy,cposz;
  static double w_model[16];
  static double w_proj[16];
  static int w_view[4];
  glGetDoublev(GL_MODELVIEW_MATRIX,w_model);
  glGetDoublev(GL_PROJECTION_MATRIX,w_proj);
  glGetIntegerv(GL_VIEWPORT,w_view);

  gluUnProject(0.0,0.0,0.0,w_model,w_proj,w_view,&cposx,&cposy,&cposz);
  return TVector(-cposx,-cposy,-cposz);
}
*/

/**
****************************************************************************************************
@brief Add texture to scene. If texture has been loaded, a texture pointer from cache is used
instead of reloading from file
@param name texture name
@param file external texture file (.tga)
@param textype texture type (can be BASE,ENV,BUMP)
@param texmode texture addition mode (can be ADD,MODULATE,DECAL,BLEND)
@param intensity texture color intensity (or bump intensity)
@param tileX count of horizontal tiles
@param tileY count of vertical tiles
@param mipmap should we generate mipmaps?
@param aniso should we use anisotropic filtering?
***************************************************************************************************/
void TScene::AddTexture(const char *name, const char *file, GLint textype, GLint texmode,
                        GLfloat intensity, GLfloat tileX, GLfloat tileY, bool mipmap, bool aniso)
{
	//material existence control
	if(materials.find(name) == materials.end())
	{
		cerr<<"WARNING (AddTexture): no material with name "<<name<<"\n";
		return;
	}
  ///find out, if texture hasn't been loaded yet
  it = tex_cache.find(file);
  GLint cache;
  ///if no match, load texture normally
  if( it == tex_cache.end() )
		cache = -1;
  ///else use existing texture
  else
		cache = it->second;
  ///then save texture ID to cache
	int texID;
	if( (texID = materials[name].AddTexture(file,textype,texmode,intensity,tileX,tileY,mipmap,aniso,cache)) == ERR)
		throw ERR;
  tex_cache[file] = texID;
	LoadScreen();	//update loading screen
}


/**
****************************************************************************************************
@brief Add cubemap texture to scene. If texture has been loaded, a texture pointer from cache is used
instead of reloading from file
@param name texture name
@param files array of 6 external texture files (.tga)
@param textype texture type (CUBEMAP)
@param texmode texture addition mode (can be ADD,MODULATE,DECAL,BLEND)
@param intensity texture color intensity (or bump intensity)
@param tileX count of horizontal tiles
@param tileY count of vertical tiles
@param aniso should we use anisotropic filtering?
***************************************************************************************************/
void TScene::AddTexture(const char *name, const char **files, GLint textype, GLint texmode,
                        GLfloat intensity, GLfloat tileX, GLfloat tileY, bool aniso)
{
	//material existence control
	if(materials.find(name) == materials.end())
	{
		cerr<<"WARNING (AddTexture): no material with name "<<name<<"\n";
		return;
	}

  //find out, if texture hasn't been loaded yet
  it = tex_cache.find(files[0]);
  GLint cache;
  ///if no match, load texture normally
  if( it == tex_cache.end() )
    cache = -1;
  ///else use existing texture
  else
    cache = it->second;
  ///then save texture ID to cache
	int texID;
	if( (texID = materials[name].AddTexture(files,textype,texmode,intensity,tileX,tileY,aniso,cache)) == ERR)
		throw ERR;
  tex_cache[files[0]] = texID;
	LoadScreen();	//update loading screen
}

/**
****************************************************************************************************
@brief Bind material(identified by mat_name) to object(identified by obj_name)
@param obj_name object name
@param mat_name material name
***************************************************************************************************/
void TScene::SetMaterial(const char* obj_name, const char *mat_name)
{
	//object existence control
	if(objects.find(obj_name) == objects.end())
	{
		cerr<<"WARNING (SetMaterial): no object with name "<<obj_name<<"\n";
		return;
	}
	//material existence control
	if(materials.find(mat_name) == materials.end())
	{
		cerr<<"WARNING (SetMaterial): no object with name "<<mat_name<<"\n";
		return;
	}
	objects[obj_name].SetMaterial(&materials[mat_name]);
}

/**
****************************************************************************************************
@brief Add 3DS object to scene. If objects has been loaded, a object pointer from cache is used
instead of reloading from file
@param name object name
@param file 3DS file with data
***************************************************************************************************/
void TScene::AddObject(const char *name, const char* file)
{
  ///find out, if object hasn't been loaded yet
  iob = obj_cache.find(file);

  ///if no match, load object normally
  if( iob == obj_cache.end() )
	{
		VBO vbo_ret = objects[name].Create(name,file,true);
		if(vbo_ret.vao == 0)
			throw ERR;
    obj_cache[file] = vbo_ret;
	}
  ///else use existing object data
  else
  {
    objects[name].SetVBOdata(iob->second);   //must be called before create!
    objects[name].Create(name,file,false);
  }
	LoadScreen();	//update loading screen
	//set sceneID
	objects[name].SetSceneID(sceneID);
}

