/**
****************************************************************************************************
****************************************************************************************************
@file: scene.h
@brief initialization and drawing the scene - declarations
****************************************************************************************************
***************************************************************************************************/
#ifndef _SCENE_H_
#define _SCENE_H_

#include "globals.h"
#include "object.h"
#include "light.h"
#include "camera.h"


/**
@class TScene
@brief containts list of objects, inits and draws 3D scene.
It has 3 main lists: objects, materials and lights. Materials are being applied to objects and
all this is lit by lights.
***************************************************************************************************/
class TScene
{
    protected:
        ///associative array with all objects
        map<string,TObject> objects;
				///iterator for objects container
        map<string,TObject>::iterator io;
        ///associative array with all materials
        map<string,TMaterial> materials;
				///iterator for materials container
        map<string,TMaterial>::iterator im;
        ///associative array with all lights
        vector<TLight> lights;
				///iterator for lights container
        vector<TLight>::iterator il;

        ///texture cache: all loaded textures are stored here; if there's request for load of already
        ///loaded texture, cache texture will be used. Cache contains source file and pointer to
        ///already created texture
        map<string,GLuint> tex_cache;
				///iterator for texture cache container
        map<string,GLuint>::iterator it;
        ///3DS objects cache - purpose is the same as texture cache
        map<string,VBO>::iterator iob;
				///iterator for object cache container
        map<string,VBO> obj_cache;


        ///2D font texture
        Texture font2D_tex;
				///Background texture
				Texture font2D_bkg;
				///Display list for font
        GLuint font2D;

        //scene parameters

				///resolution X
        GLint resx;
				///resolution Y
				GLint resy;
				///near clipping plane
        GLfloat near_p;
				///far clipping plane
				GLfloat far_p;
				///field of view
				GLfloat fovy;

				///scene camera
        TCamera cam;
				///do we use custom camera?
        bool custom_cam;
				///do we use geometry shader for dynamic cubemapping?
        bool use_gshader_ref;
				///number of multisaples in antialiasing
        int msamples;

        ///camera view matrix
        TMatrix viewMatrix;
				///temporary matrix
				TMatrix tmpMatrix;

        //scene render to texture target (FBO) and cubemap targets

				///shall we use HDR?
        bool useHDR;
				///shall we use SSAO?
				bool useSSAO;
				///shall we use shadows?
				bool useShadows;

				///framebuffer object
        GLuint f_buffer;
				///renderbuffer for color
				GLuint r_buffer_color;
				///renderbuffer for depth
				GLuint r_buffer_depth;

				///multisampled framebuffer object
        GLuint msaa_f_buffer;
				///multisampled renderbuffer for color
				GLuint msaa_r_buffer_color;
				///multisampled renderbuffer for depth
				GLuint msaa_r_buffer_depth; 

				///render target size in X
				int RT_resX;
				///render target size in Y
				int RT_resY;					

        ///basic render textures
        GLuint render_texture;
				///texture for bloom effect
				GLuint bloom_texture;
				///texture for normal values
				GLuint normal_texture;
				///texture for blurring
				GLuint blur_texture;

				///number of items to be loaded
				int load_list;
				///actual item in list
				int load_actual;
        ///actual test name
				string testname;	

				///subscene ID - used when we need to hide some parts of the scene
				int sceneID;

    public:
        //basic constructor
        TScene();
        //free dynamic data
        ~TScene();

        //scene initialization
        bool PreInit(GLint _resx, GLint _resy, GLfloat near, GLfloat far, GLfloat fovy, int msamples, bool cust_cam = false, bool load_font = true);
        bool PostInit();
        //resize window
        void Resize(GLint _resx, GLint _resy);
        //draw scene
        void Redraw(bool delete_buffer = true);
				//draw load screen
				void LoadScreen(bool swap = true);
        //destroy scene, free dynamic data
        void Destroy(bool delete_cache = true);
				///@brief set load items count (extern objects, textures and scenes)
				void SetLoadList(int count)
				{ load_list = count; load_actual = count; }
				///@brief update load list (when loading scene)
				void UpdateLoadList(int count)
				{  load_list += count; load_actual += count; }
				///@brief set test name
				void TestName(string name)
				{ testname = "Current test: "; testname += name; }

				//load whole scene from 3DS file
				void LoadScene(const char* file, bool load_materials = true, bool load_lights = true, string name_space = "");

				///Change scene ID. All objects will from now belong to this ID. Only one sceneID section can be active at time
				void ChangeSceneID(int id)
				{  sceneID = id; }

        /////////////////////////////////////////// CAMERA ////////////////////////////////////////

        ///@brief Move camera to new position(relative)
        void MoveCamera(GLfloat wx, GLfloat wy, GLfloat wz)
            { viewMatrix = cam.Move(wx,wy,wz); }
        ///@brief Move camera to new position(absolute)
        void MoveCameraAbs(GLfloat wx, GLfloat wy, GLfloat wz)
            { viewMatrix = cam.MoveAbs(wx,wy,wz); }
				///@brief Camera look
				void LookCameraAt(GLfloat wx, GLfloat wy, GLfloat wz)
					 { viewMatrix = cam.LookAt(wx,wy,wz); }

        ///@brief Rotate camera around axis(A_X, A_Y or A_Z) by angle(relative)
        void RotateCamera(GLfloat angle, GLint axis)
            { viewMatrix =  cam.Rotate(angle,axis); }

        ///@brief Rotate camera around axis(A_X, A_Y or A_Z) by angle(absolute)
        void RotateCameraAbs(GLfloat angle, GLint axis)
            { viewMatrix =  cam.RotateAbs(angle,axis); }

        ///@brief Switch between custom and scene camera
        void SwitchCamera()
            { custom_cam = !custom_cam; }

        ///@brief Print out camera position
        void PrintCamera()
				{ cout<<"POS: "<<cam.GetPos().x<<","<<cam.GetPos().y<<","<<cam.GetPos().z<<"\n"
						  <<"ROT: "<<cam.GetRot().x<<","<<cam.GetRot().y<<","<<cam.GetRot().z<<"\n"; }
        ///@brief Get screen-space camera position
        TVector GetCameraPos()
            { return cam.GetPos(); }
        //Get world camera position after transformations
        //TVector GetWorldCameraPos();



        /////////////////////////////////////////// LIGHTS ////////////////////////////////////////

        ///@brief Add light from TLight() object. New light is pushed to the end of the list
        void AddLight(GLint _lights, TVector amb, TVector diff, TVector spec, TVector lpos, float radius = 1000.0)
        { lights.push_back( TLight(_lights, amb, diff, spec, lpos, radius) ); }

        ///@brief Remove light identified by number
        ///@todo remove selected light only, not whole list
        void RemoveLight(/*int lightnum*/)
				{ lights.clear(); }

        ///@brief Move light identified by number to new absolute position (see TLight::Move() )
        void MoveLight(GLint light, TVector w)
        {
					if(light < 0 || (unsigned)light > lights.size()) cerr<<"WARNING: no light with index "<<light<<"\n";
					else lights[light].Move(w);
				}

        ///@brief Change color of light(identified by index)
        ///component(GL_AMBIENT,GL_DIFFUSE or GL_SPECULAR)  (see TLight::ChangeColor() )
        void ChangeLightColor(GLint light, GLint component, TVector color)
        {
					if(light < 0 || (unsigned)light > lights.size()) cerr<<"WARNING: no light with index "<<light<<"\n";
					else lights[light].ChangeColor(component,color);
				}

        ///@brief Return number of lights in scene
        int GetLightCount()
        { return lights.size(); }

        ///@brief Return light position
        TVector GetLightPos(int light)
        {
					if(light < 0 || (unsigned)light > lights.size()) { cerr<<"WARNING: no light with index "<<light<<"\n"; return TVector(); }
					else return lights[light].GetPos();
				}

				///@brief Update light count in shaders
				void UpdateLightCount(int count = -1)
				{
					int lcount = lights.size();
					if(count != -1 && count <= 8)
						lcount = count;
					for(im = materials.begin(); im != materials.end(); im++)
						im->second.SetLights(lcount);
				}

				///@brief set light radius
				void SetLightRadius(int light, float radius)
        {
					if(light < 0 || (unsigned)light > lights.size()) cerr<<"WARNING: no light with index "<<light<<"\n";
					else lights[light].SetRadius(radius);
				}

        /////////////////////////////////////////// OBJECTS ////////////////////////////////////////

        ///@brief Add new object (see TObject() into list. Index is object name
        void AddObject(const char *name, int primitive, GLfloat size = 0.0, GLfloat height = 0.0, GLint sliceX = 1, GLint sliceY = 1)
        {
					objects[name].Create(name, primitive, size, height, sliceX, sliceY);
					LoadScreen(); //update loading screen
					objects[name].SetSceneID(sceneID);
				}
				///@brief add new object as instance from existing object
				void AddObjectInstance(const char *ref_name, const char *inst_name)
				{
					if(objects.find(ref_name) == objects.end()) 
						cerr<<"WARNING (AddObjectInstance): no reference object with name "<<ref_name<<"\n";
					else
					{
						objects[inst_name].CreateInstance(objects[ref_name]);
						objects[inst_name].SetSceneID(sceneID);
					}
				}
        //add new object from external file
        void AddObject(const char *name, const char* file);

        ///@brief Move object identified by name to new position(relative) (see TObject::Move() )
        void MoveObj(const char* name, GLfloat wx, GLfloat wy, GLfloat wz)
        { objects[name].Move(wx,wy,wz); }
        ///@brief Move object identified by name to new position(absolute) (see TObject::MoveAbs() )
        void MoveObjAbs(const char* name, GLfloat wx, GLfloat wy, GLfloat wz)
        { objects[name].MoveAbs(wx,wy,wz); }

        ///@brief Rotate object identified by name around axis(can be A_X, A_Y, A_Z) by angle(relative)
        ///(see TObject::Rotate() )
        void RotateObj(const char* name, GLfloat angle, GLint axis)
        { objects[name].Rotate(angle,axis); }
        ///@brief Rotate object identified by name around axis(can be A_X, A_Y, A_Z) by angle(absolute)
        ///(see TObject::RotateAbs() )
        void RotateObjAbs(const char* name, GLfloat angle, GLint axis)
        { objects[name].RotateAbs(angle,axis); }

        ///@brief Resize object identified by name according to resize factor
        void ResizeObj(const char* name, GLfloat sx, GLfloat sy, GLfloat sz)
        { objects[name].Resize(sx,sy,sz); }
				///@brief Return object's vertex buffer ID
				GLint GetVertexBuffer(const char* name)
				{ return objects[name].GetVertexBuffer(); }


        ///@brief Get object position
        TVector GetObjPosition(const char *name)
        { return objects[name].GetPosition(); }

        ///@brief toggle object drawing
        void DrawObject(const char* obj_name, bool flag)
        { objects[obj_name].DrawObject(flag); }

				///@brief Set count of object instances
				void SetGInstances(const char* obj_name, int count)
				{ objects[obj_name].SetGInstances(count); }



        /////////////////////////////////////// MATERIALS&TEXTURES /////////////////////////////////

        ///@brief Add new material(see TMaterial()) to list
        void AddMaterial(const char* _name, TVector amb = black, TVector diff = silver, TVector spec = white,
                  GLfloat shin = 64.0, GLfloat reflect = 0.0, GLfloat transp = 0.0, GLint lm = PHONG)
        {
					materials[_name] = TMaterial(_name, amb, diff, spec, shin, reflect, transp, lm);
					materials[_name].SetSceneID(sceneID);
					LoadScreen(); //update loading screen
				}
        //add texture
        void AddTexture(const char *name, const char *file, GLint textype = BASE, GLint texmode = MODULATE,
                        GLfloat intensity = 1.0, GLfloat tileX = 1.0, GLfloat tileY = 1.0, bool mipmap = true, bool aniso = false);

        //add cubemap
        void AddTexture(const char *name, const char **files, GLint textype = CUBEMAP, GLint texmode = MODULATE,
                        GLfloat intensity = 1.0, GLfloat tileX = 1.0, GLfloat tileY = 1.0, bool aniso = false);

        //bind material to object
        void SetMaterial(const char* obj_name, const char *mat_name);

        ///@brief Load custom shader from file. Shader can be then applied to any material
        ///(see TMaterial::CustomShader() ). Geometry shader is optional, likewise shader defines
        void CustomShader(const char *name, const char* vert_source, const char* geom_source, const char* frag_source,
                          const char *vert_defines = NULL, const char *frag_defines = NULL, const char *geom_defines = NULL)
        {
					if(materials.find(name) == materials.end())
						cerr<<"WARNING (CustomShader): no material with name "<<name<<"\n";
					else
						if(!materials[name].CustomShader(vert_source, geom_source, frag_source,vert_defines, frag_defines, geom_defines))
							throw ERR;
				}

        ///@brief Set float uniform variable in material shader (see TMaterial::SetUniform() )
		    void SetUniform(const char* m_name, const char* v_name, float value)
			  { materials[m_name].SetUniform(v_name,value); }
				///@brief Set double uniform variable in material shader (see TMaterial::SetUniform() )
		    void SetUniform(const char* m_name, const char* v_name, double value)
			  { materials[m_name].SetUniform(v_name,value); }
				///@brief Set int uniform variable in material shader (see TMaterial::SetUniform() )
		    void SetUniform(const char* m_name, const char* v_name, int value)
			  { materials[m_name].SetUniform(v_name,value); }
        ///@brief Set array of uniform variables (eg. vector) in material shader (see TMaterial::SetUniforms() )
		    void SetUniforms(const char* m_name, const char* v_name, float *value, int values)
			  { materials[m_name].SetUniforms(v_name,value,values); }

				///@brief Force regeneration of all materials
				void BakeAllMaterials()
				{ for(im = materials.begin(); im != materials.end(); im++) im->second.BakeMaterial(); }



        ////////////////////////////////////////////SHADOWS ///////////////////////////////////////

        //create shadow map for selected light
        bool CreateShadowMap(vector<TLight>::iterator ii);
        //render shadow map
        void RenderShadowMap(TLight l);

        ///@brief Set shadow parameters(shadow size and intensity) for selected light (by index)
        ///(see TLight::SetShadow()
        void SetShadow(GLint lightNum, GLint shadow_size = 2048, GLfloat _shadow_intensity = 0.5, bool shadow = true)
        {
					if(lightNum < 0 || (unsigned)lightNum > lights.size()) cerr<<"WARNING: no light with index "<<lightNum<<"\n";
					else lights[lightNum].SetShadow(shadow_size,_shadow_intensity,shadow);
				}

        ///@brief Enable/disable shadow casting by selected object (by name) (see TObject::CastShadow() )
        void CastShadow(const char *obj_name, bool flag)
        { objects[obj_name].CastShadow(flag); }

        ///@brief Enable/disable shadow receiving for selected material (by name) (see TMaterial::ReceiveShadow() )
        void ReceiveShadow(const char *mat_name, bool flag)
        { materials[mat_name].ReceiveShadow(flag); }

        ///@brief GLobally enable/disable shadows (for enable, shadows must be created for light that casts them
        void UseShadows(bool flag = true)
        { useShadows = flag; }


        ////////////////////////// RENDER TARGETS, HDR AND DYNAMIC CUBEMAPPING ////////////////////////

        //prepare texture to render into
        void CreateRenderTarget(int resX = -1, int resY = -1, GLint tex_format = GL_RGBA16F, GLenum tex_type = GL_FLOAT);
        //resize render target textures
        void ResizeRenderTarget(int resX, int resY, GLint tex_format = GL_RGBA16F, GLenum tex_type = GL_FLOAT);
        //prepare cubemap to render into
        bool CreateRenderCubeMap(TObject *obj);
        //render scene to all faces of cube map
        void RenderToCubeMap(TObject *obj);

        ///@brief set dynamic reflection cubemap size for selected object (by name)
        void SetReflectionSize(const char* obj_name, int value)
        { objects[obj_name].refl_size = value; }

        //toggle one-pass rendering to cube map using geometry shader
        void UseDynReflections(bool use_gs = false);
        ///@brief toggle use of HDR
        void UseHDR(bool flag = true)
        {  useHDR = flag; }
        ///@brief toggle use of SSAO
        void UseSSAO(bool flag = true)
        {  useSSAO = flag; }


        ///////////////////////////////////// TEXT ////////////////////////////////////////////////

        //Draw text on screen
        void DrawScreenText(const char *s, float x, float y, float size = 1.0);
        void BuildFont();

        ////////////////////////////////////// MISCELLANEOUS //////////////////////////////////////

        ///@brief Get screen width
        int GetResX() { return resx; }
        ///@brief Get screen height
        int GetResY() { return resy; }
};
#endif
