/**
****************************************************************************************************
****************************************************************************************************
@file: shadow.cpp
@brief shadow mapping algorithms - inspired by http://www.paulsprojects.net/tutorials/smt/smt.html
****************************************************************************************************
***************************************************************************************************/
#include "scene.h"

///Distance from light
float lpos_dist;
///Temporary variable
float tmp;
///Position of light vector
TVector lpos;
///Position of eye vector
TVector eye;
///View matrix
TMatrix view;
///Eye matrix
TMatrix eye_matrix(
	1.0, 0.0, 0.0, 0.0,
	0.0, 1.0, 0.0, 0.0,
	0.0, 0.0, 1.0, 0.0,
	1.0, 1.0, 1.0, 1.0 );


/**
****************************************************************************************************
@brief Creates shadow map for selected light. Shadow map parameters are stored in TLight object
@param ii iterator to lights list
@return success/fail of shadow creation
****************************************************************************************************/
bool TScene::CreateShadowMap(vector<TLight>::iterator ii)
{
  cout<<"Shadow Map \""<<ii->GetShadowCache()<<"\": "<<ii->ShadowSize()<<"x"<<ii->ShadowSize()<<endl;

  ///1. generate depth texture for shadow map with desired dimensions for current light
  
  glGenTextures(1, &tex_cache[ii->GetShadowCache()]);
  glBindTexture(GL_TEXTURE_2D, tex_cache[ii->GetShadowCache()]);
  glTexImage2D(	GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT, ii->ShadowSize(), ii->ShadowSize(), 0,
                GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, NULL);
  
  ///2. set depth texture parameters
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

  ///3. Enable shadow comparison in hardware
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
  //Shadow comparison should be true (ie not in shadow) if r<=texture
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL);
  //Shadow comparison should generate an INTENSITY result
  glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE, GL_INTENSITY);

  ///4. create framebuffer object and bind shadow texture into it
  glGenFramebuffers(1, &ii->fbo);
  glBindFramebufferEXT(GL_FRAMEBUFFER, ii->fbo);
  //add shadow texture
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, tex_cache[ii->GetShadowCache()], 0);
  glDrawBuffer(GL_NONE); glReadBuffer(GL_NONE);
  //check FBO creation
  if( glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
  {
    ShowMessage("ERROR: FBO creation for shadow map failed!",false);
    return false;
  }
  glBindFramebufferEXT(GL_FRAMEBUFFER, 0);

  glLoadIdentity();
  gluPerspective(fovy, (GLfloat)resx/resy, near_p, far_p);
  glGetFloatv(GL_MODELVIEW_MATRIX, ii->cameraProjectionMatrix);

  return true;
}


/**
****************************************************************************************************
@brief Draws scene to shadow map
@param l current light
@return success/fail of shadow creation
****************************************************************************************************/
void TScene::RenderShadowMap(TLight l)
{
  //no shadows - clean shadow texture
  if(!useShadows)
  {
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, l.fbo);
    glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_TEXTURE_2D, tex_cache[l.GetShadowCache()], 0);
    glDrawBuffer(GL_NONE); 
    glClear(GL_DEPTH_BUFFER_BIT);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    return;
  }  
    
  ///1. set and get light projection and light view matrix
  ///save camera view matrix
  glGetFloatv(GL_MODELVIEW_MATRIX, l.cameraViewMatrix);

  ///get light projection matrix
  glLoadIdentity();
  gluPerspective(120.0f, 1.0f, 1.0f, 1000.0f);
  glGetFloatv(GL_MODELVIEW_MATRIX, l.lightProjectionMatrix);

  ///get light view matrix
  glLoadIdentity();
  lpos = l.GetPos();
/*
	//Simple correction of shadow map:
	//shorten light vector if light is too far (gets better shadow quality)
	lpos_dist = lpos.Lenght();
	if(lpos_dist > 20.0)
		lpos = lpos * 0.5;
*/
  gluLookAt(lpos.x, lpos.y, lpos.z,
            0.0f, 0.0f, 0.0f,
						//cam.GetPos().x, cam.GetPos().y, cam.GetPos().z,
            0.0f, 1.0f, 0.0f);
  glGetFloatv(GL_MODELVIEW_MATRIX, l.lightViewMatrix);
  
  ///2.bind framebuffer to draw into and draw scene from light point of view
  glBindFramebufferEXT(GL_FRAMEBUFFER, l.fbo);
  //"blank" shadow texture
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_ATTACHMENT, GL_TEXTURE_2D, tex_cache[l.GetShadowCache()], 0);
  glDrawBuffer(GL_NONE); 
  
  glMatrixMode(GL_PROJECTION); 
	glLoadMatrixf(l.lightProjectionMatrix);
  glMatrixMode(GL_MODELVIEW); 
	glLoadMatrixf(l.lightViewMatrix);

  ///draw only back faces of polygons
  glCullFace(GL_FRONT);
  glColorMask(0, 0, 0, 0);      //disable colorbuffer write
  glClear(GL_DEPTH_BUFFER_BIT);

  //set viewport, draw scene
  glViewport(0,0,l.ShadowSize() - 1, l.ShadowSize() - 1);

  ///enable polygon offset to avoid z-fighting
  glPolygonOffset(4,1);
  glEnable(GL_POLYGON_OFFSET_FILL);

  ///All scene is drawn(TObject::Draw()) without materials and lighting (only depth values and alpha tests are needed)
  for(io = objects.begin(); io != objects.end(); io++)
		if(io->second.GetSceneID() == sceneID)
			io->second.Draw(l.lightViewMatrix,false);

  //finish
  glPolygonOffset(0,0);
  glDisable(GL_POLYGON_OFFSET_FILL);

  glCullFace(GL_BACK);
  glColorMask(1, 1, 1, 1);
  glBindFramebufferEXT(GL_FRAMEBUFFER, 0);
  glClear(GL_DEPTH_BUFFER_BIT);

  //*************************
  ///3. Apply shadow map to scene - load camera projection and view matrix

  glMatrixMode(GL_PROJECTION); 
	glLoadMatrixf(l.cameraProjectionMatrix);
  glMatrixMode(GL_MODELVIEW); 
	glLoadMatrixf(l.cameraViewMatrix);
  glViewport(0, 0, resx, resy);         //reset viewport

  ///Calculate and inverse view matrix
  glGetFloatv(GL_MODELVIEW_MATRIX, view);

  //the inverse of the 3x3 rotation part is just the transpose
  for(int i=0;i<3;i++)
		for(int j=0;j<i;j++)
      std::swap(*view[4 * i + j], *view[4 * j + i]);

  //the inverse of the translation part is just the negative
  for(int i=0; i<3; i++)
		eye[i] = -*view[12 + i];
  eye[3] = 0;

  //and clear the translation part so that view is just a rotation matrix
  for(int i=0;i<3;i++) {
		*view[4 * i + 3] = 0;
		*view[12 + i] = 0;
  }

  ///4. Calculate texture shadow projection matrix by multiplying bias, view and light matrix
	*eye_matrix[12] = eye[0];
	*eye_matrix[13] = eye[1];
	*eye_matrix[14] = eye[2];
	l.textureMatrix = l.biasMatrix * l.lightProjectionMatrix * l.lightViewMatrix * view * eye_matrix;

  ///5.set created shadow texture matrix for every material (TMaterial::SetShadowMatrix())
  for(im = materials.begin(); im != materials.end(); im++)
		if(im->second.GetSceneID() == sceneID)
			im->second.SetShadowMatrix(l.textureMatrix, l.GetOrd() );
}
