/**
****************************************************************************************************
****************************************************************************************************
@file: texture.cpp
@brief load TGA textures from file and converts them to suitable OpenGL format. TGA files can be
* compressed or uncompressed
Loading TGA texture code adapted from: http://nehe.gamedev.net/data/lessons/lesson.asp?lesson=33
****************************************************************************************************
***************************************************************************************************/
#include "texture.h"

/**
****************************************************************************************************
@brief reset basic values (null name and texture data)
****************************************************************************************************/
Texture::Texture()
{
  texname = "null";
  textype = BASE;
  imageData = NULL;
  texID = width = height = bpp = 0;
  texmode = MODULATE;
  tileX = tileY = 1.0;
  texLoc = tileXLoc = tileYLoc;
}

/**
****************************************************************************************************
@brief free all texture data by detaching texture
@todo texture destruction might be called after GL context is destroyed, so it has no effect
****************************************************************************************************/
Texture::~Texture()
{
  glDeleteTextures(1,&texID);
}

/**
****************************************************************************************************
@brief load TGA texture as texture image. If texture has been loaded, a texture pointer from cache is used
instead of reloading from file
@param _texname texture name
@param _textype texture type (can be BASE,ENV,BUMP)
@param filename external texture file (.tga)
@param _texmode texture addition mode (can be ADD,MODULATE,DECAL,BLEND)
@param _intensity texture color intensity (or bump intensity)
@param _tileX count of horizontal tiles
@param _tileY count of vertical tiles
@param mipmap sholud we generate mipmaps for texture?
@param aniso should we use anisotropic filtering?
@param cache whether we should use cached texture image or load new
@return new texture ID
****************************************************************************************************/
GLint Texture::Load(const char *_texname, int _textype, const char *filename, int _texmode,
                    GLfloat _intensity, GLfloat _tileX, GLfloat _tileY, bool mipmap, bool aniso, GLint cache)
{
  ///if texture is loaded in cache, copy only pointer to texture data
  if(cache != -1)
  {
    texID = cache;
  }
  else      //load texture from file
  {
    if(!LoadTGA(filename))
      return ERR;

    //texture generation
    glGenTextures(1, &texID);
    glBindTexture(GL_TEXTURE_2D, texID);

    //texture with anisotropic filtering
    if(aniso)
    {
      //find out, if GFX supports aniso filtering
      if(!GLEW_EXT_texture_filter_anisotropic)
      {
        cout<<"Anisotropic filtering not supported. Using linear instead.\n";
      }
      else
      {
        float maxAnisotropy;
        //find out maximum supported anisotropy
        glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &maxAnisotropy);
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAX_ANISOTROPY_EXT, maxAnisotropy);
      }
    }
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    //mip-mapped texture (if set)
    if(mipmap)
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
    else
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);

		if(_textype == BUMP)	//don't compress bump maps
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, imageData);
		else								//compress color maps
			glTexImage2D(GL_TEXTURE_2D, 0, GL_COMPRESSED_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, imageData);

    if(mipmap)
#ifdef GL3   
		glGenerateMipmap(GL_TEXTURE_2D);
#else
		gluBuild2DMipmaps(GL_TEXTURE_2D, 3, width,height, GL_RGB, GL_UNSIGNED_BYTE,imageData);
#endif

    delete [] imageData;      //don't need image data after texture was created
  }

  //texture mode and type
  texname = _texname;
  textype = _textype;
  texmode = _texmode;
  tileX = _tileX;
  tileY = _tileY;
  intensity = _intensity;

  //cout<<"Texture \""<<texname<<"\" creation OK\n";

  return texID;
}


/**
****************************************************************************************************
@brief load 6 TGA texture as cube map. If texture has been loaded, a texture pointer from cache is used
instead of reloading from file
@param _texname texture name
@param _textype texture type (can be BASE,ENV,BUMP)
@param cube_files array of strings with filenames containing every side of cube map (.tga)
@param _texmode texture addition mode (can be ADD,MODULATE,DECAL,BLEND)
@param _intensity texture color intensity (or bump intensity)
@param _tileX count of horizontal tiles
@param _tileY count of vertical tiles
@param aniso should we use anisotropic filtering?
@param cache whether we should use cached texture image or load new
@return new texture ID
****************************************************************************************************/

GLint Texture::Load(const char *_texname, int _textype,
                    const char **cube_files, int _texmode, GLfloat _intensity, GLfloat _tileX, GLfloat _tileY, bool aniso, GLint cache)
{
  ///if texture is loaded in cache, copy only pointer to texture data
  if(cache != -1)
  {
    texID = cache;
  }
  else      //load texture from file
  {
    //texture generation
    glGenTextures(1, &texID);
    glBindTexture(GL_TEXTURE_CUBE_MAP, texID);

    //load 6 images for cube map
    for(int i = 0; i < 6; i++)
    {
      if(!LoadTGA(cube_files[i]))
        return ERR;

      //texture with anisotropic filtering
      if(aniso)
      {
        //find out, if GFX supports aniso filtering
        if(!GLEW_EXT_texture_filter_anisotropic)
        {
          cout<<"Anisotropic filtering not supported. Using linear instead.\n";
        }
        else
        {
          float maxAnisotropy;
          //find out maximum supported anisotropy
          glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &maxAnisotropy);
          glTexParameterf(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, GL_TEXTURE_MAX_ANISOTROPY_EXT, maxAnisotropy);
        }
      }

      //mip-mapped texture
      glTexParameteri(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, GL_TEXTURE_MAG_FILTER,GL_LINEAR);
      glTexParameteri(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
      glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 0, GL_COMPRESSED_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, imageData);
#ifdef GL3      
	  glGenerateMipmap(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i);
#else      
	  gluBuild2DMipmaps(GL_TEXTURE_CUBE_MAP_POSITIVE_X + i, 3, width,height, GL_RGB, GL_UNSIGNED_BYTE,imageData);
#endif

      delete [] imageData;      //don't need image data after texture was created
    }
  }

  //texture mode and type
  texname = _texname;
  textype = _textype;
  texmode = _texmode;
  tileX = _tileX;
  tileY = _tileY;
  intensity = _intensity;

  //cout<<"Texture \""<<texname<<"\" creation OK\n";

  return texID;
}



/**
****************************************************************************************************
@brief load and decode TGA image from file
@param filename file with image data
@return success/fail of texture loading
****************************************************************************************************/
bool Texture::LoadTGA(const char* filename)
{
	FILE* fTGA;											//file
	int type;

	GLubyte tgaheader[12];                             //TGA header

  //headers: for compressed and for uncompressed images
  GLubyte uTGAcompare[12] = { 0,0, 2,0,0,0,0,0,0,0,0,0 };
  GLubyte cTGAcompare[12] = { 0,0,10,0,0,0,0,0,0,0,0,0 };

	//open file
	fTGA = fopen(filename, "rb");
	if(fTGA == NULL)
	{
    string err = "Can't open texture file ";
    err += filename;
    err += " !";
		ShowMessage(err.c_str(), false);
		return false;
	}

	cout<<"Loading image "<<filename<<"...";

	///read and check file header
	if(fread(&tgaheader, 12, 1, fTGA) == 0)
	{
		ShowMessage("Can't read image header!");
		if(fTGA != NULL) fclose(fTGA);
		return false;
	}

	///find out from image, whether data are compressed or not
	if(memcmp(uTGAcompare, &tgaheader, sizeof(tgaheader)) == 0)
	  type = UNCOMPRESSED;
	else if(memcmp(cTGAcompare, &tgaheader, sizeof(tgaheader)) == 0)
	  type = COMPRESSED;
	//else error
	else
	{
		ShowMessage("Unknown TGA image type!");
		fclose(fTGA);
		return false;
	}

	//read further bytes of header
	if(fread(tgaheader, 6, 1, fTGA) == 0)
	{
		ShowMessage("Can't read image header!");
		if(fTGA != NULL) fclose(fTGA);
		return false;
	}

	///extract width, height and bpp of image
	width = tgaheader[1] * 256 + tgaheader[0];
	height = tgaheader[3] * 256 + tgaheader[2];
	bpp = tgaheader[4];

  cout<<width<<"x"<<height<<"x"<<bpp<<"...";

	//check values
	if((width <= 0) || (height <= 0) || ((bpp != 24) && (bpp != 32)) )
	{
		ShowMessage("Unknown TGA image type!");
		if(fTGA != NULL) fclose(fTGA);
		return false;
	}
/*
	//set color depth
	int gl_textype;
	if(bpp == 24)
		gl_textype = GL_RGB;
	else
		gl_textype = GL_RGBA;
*/
	///allocate memory for image data
	unsigned bytesPerPixel = (bpp / 8);
	unsigned imageSize = (bytesPerPixel * width * height);
	imageData = new GLubyte[imageSize];
	if(imageData == NULL)
	{
		ShowMessage("Can't allocate image data!");
		fclose(fTGA);
		return false;
	}

	///load from file according to compression
	if(type == UNCOMPRESSED)
	{
	  ///uncompressed image - read data directly from file
	  if(fread(imageData, 1, imageSize, fTGA) != imageSize)
	  {
		  ShowMessage("Can't read image data!");
		  if(imageData != NULL) delete [] imageData;
		  fclose(fTGA);
		  return false;
	  }
	  //conversion from BGR to RGB
	  for(unsigned cswap = 0; cswap < imageSize; cswap += bytesPerPixel)
		  imageData[cswap] ^= imageData[cswap+2] ^= imageData[cswap] ^= imageData[cswap+2];
	}

	///compressed image - decode using RLE compression
	else if(type == COMPRESSED)
	{
		GLuint pixelcount = height * width;			//pixel count
		GLuint currentpixel = 0;					//actual pixel
		GLuint currentbyte = 0;						//actual byte

		//allocate memory for image pixel
		GLubyte* colorbuffer = new GLubyte[bytesPerPixel];
		if(colorbuffer == NULL)
		{
			ShowMessage("Can't allocate color buffer data!");
			fclose(fTGA);
			return false;
		}

		do
		{
      //find out header type: RLE or RAW
			GLubyte chunkheader = 0;
			if(fread(&chunkheader, sizeof(GLubyte), 1, fTGA) == 0)
			{
				ShowMessage("Can't read RLE header!");
				if(fTGA != NULL) fclose(fTGA);
				if(imageData != NULL) delete [] imageData;
				if(colorbuffer != NULL) delete [] colorbuffer;
				return false;
			}

			//1. RAW part of file
			if(chunkheader < 128)
			{
				chunkheader++;				//count pixels before other header
				//nactavame kazdy pixel
				for(short counter = 0; counter < chunkheader; counter++)
				{
					if(fread(colorbuffer, 1, bytesPerPixel, fTGA) != bytesPerPixel)
					{
						ShowMessage("Can't read image data!");
						if(fTGA != NULL) fclose(fTGA);
						if(colorbuffer != NULL) delete [] colorbuffer;
						if(imageData != NULL) delete [] imageData;
						return false;
					}
					//write to memory, conversion to RGB
					imageData[currentbyte] = colorbuffer[2];
					imageData[currentbyte + 1] = colorbuffer[1];
					imageData[currentbyte + 2] = colorbuffer[0];
					///@todo alpha channel
					//if(bytesPerPixel == 4) imageData[currentbyte + 3] = colorbuffer[3];
					//move to next byte
					currentbyte += bytesPerPixel;
					currentpixel++;

					//check pixel count, if it's according to image dimensions
					if(currentpixel > pixelcount)
					{
						ShowMessage("Wrong image format - too many pixels.");
						if(fTGA != NULL) fclose(fTGA);
						if(colorbuffer != NULL) delete [] colorbuffer;
						if(imageData != NULL) delete [] imageData;
						return false;
					}
				}
			}
			//2. RLE part of image
			else
			{
				chunkheader -= 127;						//pixel count in section
				//read pixel after pixel
				if(fread(colorbuffer, 1, bytesPerPixel, fTGA) != bytesPerPixel)
				{
					ShowMessage("Can't read image data!");
					if(fTGA != NULL) fclose(fTGA);
					if(colorbuffer != NULL) delete [] colorbuffer;
					if(imageData != NULL) delete [] imageData;
					return false;
				}
				//pixel copy
				for(short counter = 0; counter < chunkheader; counter++)
				{
					//write to memory, conversion to RGB
					imageData[currentbyte] = colorbuffer[2];
					imageData[currentbyte + 1] = colorbuffer[1];
					imageData[currentbyte + 2] = colorbuffer[0];
					//TODO: alpha channel
					//if(bytesPerPixel == 4) imageData[currentbyte + 3] = colorbuffer[3];
					//move to next byte
					currentbyte += bytesPerPixel;
					currentpixel++;

					//check pixel count, if it's according to image dimensions
					if(currentpixel > pixelcount)
					{
						ShowMessage("Wrong image format - too many pixels.");
						if(fTGA != NULL) fclose(fTGA);
						if(colorbuffer != NULL) delete [] colorbuffer;
						if(imageData != NULL) delete [] imageData;
						return false;
					}
				}
			}
		}while(currentpixel < pixelcount);

		//clear up memory
		if(colorbuffer != NULL) delete [] colorbuffer;
	}

	//done
	fclose(fTGA);
  cout<<"OK\n";
	return true;
}


/**
****************************************************************************************************
@brief Activates texture map for use by shader
@param tex_unit multitexture unit used for texture application
@param set_uniforms shall we also set uniform locations to shader?
****************************************************************************************************/
void Texture::ActivateTexture(GLint tex_unit, bool set_uniforms)
{
  ///1. set uniform values in shader (tiles, texture unit, texture matrix and intensity)
	if(set_uniforms)
	{
		glUniform1f(tileXLoc, tileX);
		glUniform1f(tileYLoc, tileY);
		glUniform1f(intensityLoc, intensity);
	}
	//texture location and shadow texture matrix must be updated regularly
	glUniform1i(texLoc, tex_unit);
	if(texMatrixloc != -1)
		glUniformMatrix4fv(texMatrixloc, 1, GL_FALSE, texMatrix);

  ///2. activate and bind texture
  glActiveTexture(GL_TEXTURE0 + tex_unit);
  //2D or cube map?
  if(textype == CUBEMAP || textype == CUBEMAP_ENV)
		glBindTexture(GL_TEXTURE_CUBE_MAP, texID);
  else
		glBindTexture(GL_TEXTURE_2D, texID);
}

/**
****************************************************************************************************
@brief Gets uniform variables froms shader
@param shader handle to shader to bound with texture
****************************************************************************************************/
void Texture::GetUniforms(GLuint shader)
{
	if(HasTiles())
	{
		string tileX_str = texname + "_tileX";
		string tileY_str = texname + "_tileY";
		tileXLoc = glGetUniformLocation(shader,tileX_str.c_str() );
		tileYLoc = glGetUniformLocation(shader,tileY_str.c_str() );
	}

	///1. to avoid mismatch variable names, use texture name as variable prefix
  string texMatrix_str = texname + "_texMatrix";
  string intensity_str = texname + "_intensity";

  ///2. get uniforms location
  texLoc = glGetUniformLocation(shader,texname.c_str());
  texMatrixloc = glGetUniformLocation(shader,texMatrix_str.c_str() );
  intensityLoc = glGetUniformLocation(shader,intensity_str.c_str() );
}

