/**
****************************************************************************************************
****************************************************************************************************
@file: texture.h
@brief load TGA textures from file and converts them to suitable OpenGL format. TGA files can be
* compressed or uncompressed
****************************************************************************************************
***************************************************************************************************/
#ifndef _TEXTURE_H_
#define _TEXTURE_H_
#include "mat.h"

///Two possible types of TGA image
enum TGAtypes{COMPRESSED,UNCOMPRESSED};

///@class Texture 
///@brief holds texture parameters and contains functions to load texture from external file
///Textures are connected to shaders via uniform variables
class Texture
{
	private:
    int textype;                  //texture type
    string texname;               //texture name
        
    GLubyte *imageData;           //image data
    GLuint width, height, bpp;    //width, height
    GLuint texID;                 //texture ID
    int texmode;
    GLfloat tileX,tileY;          //texture tiles
    //shader uniform variables
    GLint texLoc, tileXLoc, tileYLoc, texMatrixloc, intensityLoc;
    //texture matrix for shadow map
    TMatrix texMatrix;

  public:    
    ///texture intensity
    GLfloat intensity;
    
    Texture();
    ~Texture();
    
    ///@brief set texture name
    void SetName(string name) 
		{ texname = name; }
		///@brief get texture name
		string GetName()
		{ return texname; }
    
    //create texture from TGA file
    GLint Load(const char *_texname, int _textype, const char *filename, int _texmode, 
               GLfloat _intensity, GLfloat _tileX, GLfloat _tileY, bool mipmap, bool aniso, GLint cache);

    //create cubemap fro six TGA files
    GLint Load(const char *_texname, int _textype, const char **cube_files,
               int _texmode, GLfloat _intensity, GLfloat _tileX, GLfloat _tileY, bool aniso, GLint cache);

    //load TGA texture from file
    bool LoadTGA(const char *filename);

    ///@brief do we have image data?
    bool Empty() 
    { return (texID == 0); } 
    
    //activate texture for use by shader
    void ActivateTexture(GLint tex_unit, bool set_uniforms = false);
    //get uniform variables for texture
    void GetUniforms(GLuint shader);
    
    ///@brief set texture matrix for shadow map projection
    void SetShadowMatrix(TMatrix t) 
    { texMatrix = t; }    
    ///@brief set texture intensity
    void SetIntensity(GLfloat _intensity) 
    { intensity = _intensity; }
		///@brief Set texture ID
		void SetID(GLuint id)
		{ texID = id; }
		///@brief Set texture type
		void SetType(int type)
		{ textype = type; }

		///@brief Get OpenGL texture ID
		GLuint GetID()
		{ return texID; }
		///@brief Get texture type
		int GetType()
		{ return textype; }
		///@brief Get texture mode
		int GetMode()
		{ return texmode; }
		///@brief Has texture tiles?
		bool HasTiles()
		{ return (tileX > 1 || tileY > 1); }
};

#endif
