 /******************************************************************************/
//
//  Program: gluxMark2, DIP
//  Author: Juraj Vanek, xvanek29@stud.fit.vutbr.cz
//  File: test.cpp
//  Desc: test class methods which initialize and run tests, system info
//
//*****************************************************************************/
#include "test.h"

/**
****************************************************************************************************
@brief Test class basic constructor
@param _duration duration of the test
@param _type test type
@param ps pointer to scene
****************************************************************************************************/
Test::Test(int _duration, int _type, TScene *ps)
{
  s = ps;                   //pointer to scene

  //set basic values
  avg_fps = 0.0;
  fps = 100.0;
  type = _type;
  duration = _duration;
  total_frames = 0;
  anim = 0.1f;
  stopbit = false;
  curr_setup = 1;
  splitbit = true;
  start_time = 2000000000;        //(enough time for scene to load :)

	test_score = 0;

	//loading polygon
	s->AddMaterial("mat_load_polygon",white,white,white,64.0,0.0,0.0,NONE);
	s->AddTexture("mat_load_polygon","data/load.tga",BASE);
	s->CustomShader("mat_load_polygon","data/shaders/quad.vert",NULL,"data/shaders/load_quad.frag");
	//background
	s->AddMaterial("mat_background_quad",white,white,white,64.0,0.0,0.0,NONE);
	s->AddTexture("mat_background_quad","data/background.tga",BASE);
	s->CustomShader("mat_background_quad","data/shaders/quad.vert",NULL,"data/shaders/load_quad.frag");

	//set testnames
	string screen_testname;
  switch(type)
  {
    case TEXEL_FILLRATE_SINGLE:
				screen_testname = "Texel Fillrate(Single texturing)";
        break;
    case TEXEL_FILLRATE_MULTI:
				screen_testname = "Texel Fillrate(Multi texturing)";
        break;
    case TEXEL_FILLRATE_FLOAT16:
				screen_testname = "Texel Fillrate(16-bit floating point textures)";
        break;
    case TEXEL_FILLRATE_FLOAT32:
				screen_testname = "Texel Fillrate(32-bit floating point textures)";
        break;
    case FRAG_LIGHTS:
				screen_testname = "Per-pixel point lights";
        break;
    case FRAG_NOISE:
				screen_testname = "Perlin noise";
        break;
    case FRAG_PROCEDURAL:
				screen_testname = "Procedural shaders";
        break;
    case FRAG_PARALLAX:
				screen_testname = "Parallax mapping";
        break;
    case FRAG_REFRACT:
				screen_testname = "Static reflections&refractions";
        break;
    case GEOM_TESSELATE:
				screen_testname = "Tesselation with depth map";
        break;
    case GEOM_CUBEMAP:
				screen_testname = "Single-pass dynamic cubemapping";
        break;
    case GEOM_PARTICLES:
				screen_testname = "GPU particles";
        break;
    case GEOM_INSTANCING:
				screen_testname = "Geometry shader instancing";
        break;
    case VERT_LIGHTS:
				screen_testname = "High polygon count, per-vertex point lights";
        break;
    case VERT_WAVES:
				screen_testname = "GPU generated waves";
				break;
    case VERT_DISPLACE:
				screen_testname = "Vertex displacement mapping";
        break;
    case COMPLEX_HDR:
				screen_testname = "HDR lighting";
        break;
    case COMPLEX_SSAO:
				screen_testname = "Screen space ambient occlusion";
        break;
    case COMPLEX_SHADOW:
				screen_testname = "Dynamic soft shadows";
        break;
    case COMPLEX_SSAMPLING:
				screen_testname = "Render with supersampling";
        break;
    case COMPUTE_NBODY:
				screen_testname = "N-Body simulation";
        break;
    default: break;
  }
	s->TestName(screen_testname);

	cout<<"\n---------------------------------------------\n"
			  "Initializing test "<<screen_testname<<":\n";
}


/**
****************************************************************************************************
@brief Test run, counts FPS and increment time
****************************************************************************************************/
void Test::Run()
{
  draw_scene = true;
  //splashscreen
  if(SDL_GetTicks() < start_time)
  {
		s->LoadScreen(false);
    draw_scene = false;
  }
  //else normal scene drawing
  if(draw_scene)
  {
    //partial scenes (splits)
    if(SDL_GetTicks() - split_time >= duration/splits)
    {
      splitbit = true;         //set split bit after partial scene has ended
      SaveResults();
      total_frames = 0;
      split_time = SDL_GetTicks();
      curr_setup++;

      //test end after all setups gone
      if(curr_setup > splits)
      {
        stopbit = true;
        draw_scene = false;
				cout<<"Test finished.\n";
      }
    }

    if(draw_scene)
      total_frames++;           //count frames (only when scene is drawn)
  }
}


/**
****************************************************************************************************
@brief Calculate average framerate and save results
****************************************************************************************************/
void Test::SaveResults()
{
  //Average framerate
  avg_fps = (float)total_frames/(duration/1000.0f/splits);
	cout<<"Split FPS: "<<avg_fps<<endl;
}

/**
****************************************************************************************************
@brief Save system info
****************************************************************************************************/
void SystemInfo()
{
  string os,cpuinfo,cpufreq,meminfo;
  char vram[4];

  ///1. find out informations according to OS
#ifdef _WIN_
  os = "Windows";
  ///Windows - use registers

  //CPU:
  HKEY hkey;
  LONG ret = RegOpenKeyEx(HKEY_LOCAL_MACHINE,
      "HARDWARE\\DESCRIPTION\\System\\CentralProcessor\\0",0,KEY_READ,&hkey);

  DWORD size = 0;
  ret = RegQueryValueEx(hkey,"ProcessorNameString",0,NULL,0,&size);

  char *cpustr = new TCHAR[size];
  ret = RegQueryValueEx(hkey,"ProcessorNameString",0,NULL,(LPBYTE)cpustr,&size);
  cpuinfo = cpustr;

  //CPU frequency
  DWORD freq;
  ret = RegQueryValueEx(hkey,"~MHz",0,NULL,(LPBYTE)&freq,&size);
	string fr;
  fr = num2str(freq);
  cpufreq = fr;
  RegCloseKey(hkey);

  //RAM:
  MEMORYSTATUS mem;
  mem.dwLength = sizeof(mem);
  GlobalMemoryStatus(&mem);
  string memsize;
  memsize = num2str(mem.dwTotalPhys/1048576) + " MB";
  meminfo = memsize;

  //VRAM:
  //find informations in DEVICEMAP\VIDEO
  RegOpenKeyEx(HKEY_LOCAL_MACHINE,"HARDWARE\\DEVICEMAP\\VIDEO",0,KEY_READ,&hkey);
  RegQueryValueEx(hkey,"\\Device\\Video0",0,NULL,0,&size);

  char *gfxlocR = new TCHAR[size];
  RegQueryValueEx(hkey,"\\Device\\Video0",0,NULL,(LPBYTE)gfxlocR,&size);
  //cut out beginning
  string gfxlocA = gfxlocR;
  gfxlocA.erase(0,18);

  //directly open key with GFX info
  RegOpenKeyEx(HKEY_LOCAL_MACHINE, gfxlocA.c_str(),0,KEY_READ,&hkey);
  RegQueryValueEx(hkey,"HardwareInformation.MemorySize",0,NULL,(LPBYTE)vram,&size);

#else
  os = "Linux";
  ///Linux - find out from /proc/cpuinfo and /proc/meminfo
  string tmp;
  //CPU
  ifstream fin("/proc/cpuinfo");
  while(getline(fin,tmp))
  {
    if(tmp.find("model name") != string::npos)    //CPU name
    {
      tmp.erase(0,tmp.find(':')+2);
      cpuinfo = tmp;
    }
    else if(tmp.find("cpu MHz") != string::npos)    //CPU frequency
    {
      tmp.erase(0,tmp.find(':')+2);
      cpufreq = tmp;
    }
  }
  fin.close();

  //RAM
  fin.open("/proc/meminfo");
  int memsize;
  char memchar[16];
  while(getline(fin,tmp))
  {
    if(tmp.find("MemTotal") != string::npos)
    {
      tmp.erase(0,tmp.find(':') + 8);
      sscanf(tmp.c_str(),"%d",&memsize);
      memsize /= 1024;
      sprintf(memchar,"%d MB",memsize);
      meminfo = memchar;
      break;
    }
  }
  fin.close();

  vram[3] = 0;      //TODO: VRAM in Linux
#endif

	///2. OpenGL properties: get them from glGetIntegerv()
  int props[17];
  glGetIntegerv(GL_MAX_TEXTURE_SIZE,&props[0]);
  glGetIntegerv(GL_MAX_3D_TEXTURE_SIZE,&props[1]);
	glGetIntegerv(GL_MAX_ARRAY_TEXTURE_LAYERS,&props[2]);
  if(props[2] > 16536)
    props[2] = 0;
	glGetIntegerv(GL_MAX_CUBE_MAP_TEXTURE_SIZE,&props[3]);
	glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE,&props[4]);
	glGetIntegerv(GL_MAX_DRAW_BUFFERS,&props[5]);
	glGetIntegerv(GL_MAX_SAMPLES,&props[6]);
	glGetIntegerv(GL_NUM_EXTENSIONS,&props[7]);
	glGetIntegerv(GL_MAX_ELEMENTS_INDICES,&props[8]);
	glGetIntegerv(GL_MAX_ELEMENTS_VERTICES,&props[9]);

	glGetIntegerv(GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS,&props[10]);
	glGetIntegerv(GL_MAX_VARYING_COMPONENTS,&props[11]);
	glGetIntegerv(GL_MAX_VERTEX_ATTRIBS,&props[12]);
	glGetIntegerv(GL_MAX_VERTEX_UNIFORM_COMPONENTS,&props[13]);
  if(GLEW_ARB_geometry_shader4 || GLEW_EXT_geometry_shader4 || GLEW_EXT_gpu_shader4)
  {
	  glGetIntegerv(GL_MAX_GEOMETRY_OUTPUT_VERTICES,&props[14]);
	  glGetIntegerv(GL_MAX_GEOMETRY_TOTAL_OUTPUT_COMPONENTS,&props[15]);
  }
  else
  {
    props[14] = 0;
    props[15] = 0;
  }
	glGetIntegerv(GL_MAX_FRAGMENT_UNIFORM_COMPONENTS,&props[16]);

	//Supports
	string supports[5];
	for(unsigned i=0; i<5; i++)
		supports[i] = "NO";

	if(GLEW_ARB_draw_buffers)
		supports[0] = "YES";
  if(GLEW_ARB_draw_instanced || GLEW_EXT_draw_instanced)
		supports[1] = "YES";
	if(GLEW_ARB_geometry_shader4 || GLEW_EXT_geometry_shader4 || GLEW_EXT_gpu_shader4)
		supports[2] = "YES";
	if(GLEW_ARB_texture_float)
		supports[3] = "YES";
	if(GLEW_EXT_framebuffer_multisample)
		supports[4] = "YES";

	//video RAM
	string vram_str;
	if(vram[3] < 32)
		vram_str = "N/A";
	else
		vram_str = num2str(vram[3]*16);

  ///3. save to XML
  results<< "<sysinfo>\n"
              "<system>\n"
                "<os>"<<os<<"</os>\n"
                "<cpu>"<<cpuinfo<<"</cpu>\n"
                "<cpufreq>"<<cpufreq<<"</cpufreq>\n"
                "<memory>"<<meminfo<<"</memory>\n"
              "</system>\n"

              "<gfx>\n"
                "<name>"<<glGetString(GL_RENDERER)<<"</name>\n"
                "<vendor>"<<glGetString(GL_VENDOR)<<"</vendor>\n"
                "<vram>"<<vram_str<<"</vram>\n"
                "<glversion>"<<glGetString(GL_VERSION)<<"</glversion>\n"
								"<glslversion>"<<glGetString((GL_SHADING_LANGUAGE_VERSION))<<"</glslversion>\n"
								"<extensions>"<<props[7]<<"</extensions>\n"<<

								"<hasMRT>"<<supports[0]<<"</hasMRT>"
								"<hasGI>"<<supports[1]<<"</hasGI>"
								"<hasGS>"<<supports[2]<<"</hasGS>"
								"<hasFloat>"<<supports[3]<<"</hasFloat>"
								"<hasFBOMulti>"<<supports[4]<<"</hasFBOMulti>"

                "<texsize>"<<props[0]<<"</texsize>\n"
                "<texsize3D>"<<props[1]<<"</texsize3D>\n"
                "<tex_layers>"<<props[2]<<"</tex_layers>\n"
                "<cubetexsize>"<<props[3]<<"</cubetexsize>\n"
                "<renderbufsize>"<<props[4]<<"</renderbufsize>\n"
                "<drawbuffs>"<<props[5]<<"</drawbuffs>\n"
                "<msamples>"<<props[6]<<"</msamples>\n"
                "<indices>"<<props[8]<<"</indices>\n"
                "<vertices>"<<props[9]<<"</vertices>\n"

                "<comb_texunits>"<<props[10]<<"</comb_texunits>\n"
                "<varyings>"<<props[11]<<"</varyings>\n"
                "<vertattribs>"<<props[12]<<"</vertattribs>\n"
                "<vertunifs>"<<props[13]<<"</vertunifs>\n"
                "<gsout_vertices>"<<props[14]<<"</gsout_vertices>\n"
                "<gsout_components>"<<props[15]<<"</gsout_components>\n"
                "<fragunifs>"<<props[16]<<"</fragunifs>\n"
              "</gfx>\n";

#ifdef USE_CL
	//OpenCL info - initialize device 
  cl_platform_id platform;
	cl_device_id device_id;

	const int BUFF = 1024;
  cl_char vendor_name[BUFF] = {0};
  cl_char device_name[BUFF] = {0};
	cl_device_local_mem_type local_mem_type;
  cl_ulong global_mem_size;
	cl_uint clock_frequency, max_compute_units;
	size_t max_work_group_size, max_work_item_sizes[3];

	if(clGetPlatformIDs(1, &platform, NULL) == CL_SUCCESS)
	{
		results<<"<opencl>\n<isCL>YES</isCL>";
		//find the GPU CL device
	  clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &device_id, NULL);
		//query capabilities
		clGetDeviceInfo(device_id, CL_DEVICE_VENDOR, sizeof(vendor_name), vendor_name, NULL);
		clGetDeviceInfo(device_id, CL_DEVICE_NAME, sizeof(device_name), device_name,  NULL);
		clGetDeviceInfo(device_id, CL_DEVICE_LOCAL_MEM_TYPE, sizeof(local_mem_type), &local_mem_type,  NULL);
		clGetDeviceInfo(device_id, CL_DEVICE_GLOBAL_MEM_SIZE, sizeof(global_mem_size), &global_mem_size,  NULL);
		clGetDeviceInfo(device_id, CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof(clock_frequency), &clock_frequency,  NULL);
		clGetDeviceInfo(device_id, CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(max_work_group_size), &max_work_group_size,  NULL);
		clGetDeviceInfo(device_id, CL_DEVICE_MAX_WORK_ITEM_SIZES, sizeof(max_work_item_sizes), max_work_item_sizes, NULL);
		clGetDeviceInfo(device_id, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(max_compute_units), &max_compute_units,  NULL);
		//local memory suppord
		string isLocal = "NO";
		if((int)local_mem_type == 1)
			isLocal = "YES";
		
		results<<
			"<vendor>"<<vendor_name<<"</vendor>\n"
			"<device>"<<device_name<<"</device>\n"
			"<units>"<<max_compute_units<<"</units>\n"
			"<clock>"<<clock_frequency<<" MHz</clock>\n"
			"<isLocal>"<<isLocal<<"</isLocal>\n"
			"<globMem>"<<(int)global_mem_size/(1024*1024)<<" MB</globMem>\n"
			"<workitems>"<<(int)max_work_item_sizes[0]<<"x"<<(int)max_work_item_sizes[1]<<"x"<<(int)max_work_item_sizes[2]<<"</workitems>\n"
			"<workgroup>"<<(int)max_work_group_size<<"</workgroup>";
	}
	else
		results<<"<opencl>\n<isCL>NO</isCL>";

  results<<"</opencl>\n</sysinfo>";
#else
  results<<"<opencl>\n<isCL>NO</isCL>\n</opencl>\n</sysinfo>\n";
#endif
}
