/******************************************************************************/
//
//  Program: gluxMark2, DIP
//  Author: Juraj Vanek, xvanek29@stud.fit.vutbr.cz
//  File: test_compute.cpp
//  Desc: implementation of OpenCL compute test series
//
//*****************************************************************************/
#include "test_compute.h"

#ifdef USE_CL
/**
****************************************************************************************************
@brief Compute tests initialization
@param _duration test duration
@param _type test type
@param s pointer to scene
****************************************************************************************************/
TestCompute::TestCompute(int _duration, int _type, TScene *s) : Test(_duration, _type, s)
{
  //switch according to test subtype
  if(type == COMPUTE_NBODY)
  {
	  testname = "nbody";
		weight = 1.0;
    splits = 1;
		halfbody = BODIES/2;

		s->SetLoadList(7);	//loaded items count (must be determined manually :(

		try
		{
			nbody = new NBodySim();
			if(!nbody->Init())
				throw ERR;
			if(!nbody->SetupKernel("data/kernels/NBody.cl","nbody_sim"))
				throw ERR;

			//pseudo-HDR: only 8-bit render texture
			s->CreateRenderTarget(-1,-1,GL_RGBA,GL_UNSIGNED_BYTE);
			s->UseHDR();

			//position camera
			s->MoveCamera(0.0,0.0,-150.0);
			s->RotateCamera(15.0, A_X);

			s->AddMaterial("mat",black,white,black,0.0,0.0,0.0,NONE);
			s->AddMaterial("mat2",black,TVector(1.0f,0.8f,0.6f),black,0.0,0.0,0.0,NONE);

			//sky(space :)
			s->AddObject("space",SPHERE,500.0,0.0,8,8);
			s->AddMaterial("mat_space",white,white,white,0.0,0.0,0.0,NONE);
			s->AddTexture("mat_space","data/tex/stars.tga",BASE,MODULATE,1.0,6.0,4.0);
			s->SetMaterial("space","mat_space");

			//random position, two groups
			//first group of instances
			s->AddObject("1_star0",SPHERE,0.3f,0.0f,4,4);
			s->SetMaterial("1_star0","mat");
			for(int i=0; i<BODIES/2; i++)
			{
				string name = "1_star" + num2str(i + 1);
				s->AddObjectInstance("1_star0",name.c_str());
			}
			//second group of instances
			s->AddObject("2_star0",SPHERE,0.6f,0.0f,4,4);
			s->SetMaterial("2_star0","mat2");
			for(int i=0; i<BODIES/2; i++)
			{
				string name = "2_star" + num2str(i + 1);
				s->AddObjectInstance("2_star0",name.c_str());
			}

			//General HDR settings
			s->SetUniform("mat_bloom_hdr","THRESHOLD",0.2);
			s->SetUniform("mat_tonemap","exposure",1.5);
			s->SetUniform("mat_tonemap","bloomFactor",1.0);
			s->SetUniform("mat_tonemap","brightMax",1.0);

			//blur
			s->SetUniform("mat_blur_horiz","kernel_size",15);
			s->SetUniform("mat_blur_horiz","blur_radius",0.003);
			s->SetUniform("mat_blur_vert","kernel_size",15);
			s->SetUniform("mat_blur_vert","blur_radius",0.003);
		}
		catch(int err)
		{
			stopbit = true;
			throw ERR;
			return;
		}
	}
	//done
  if(!s->PostInit())
		stopbit = true;
  InitTime();
}

//**************************************************************************************************
//Test deinitialization, reset scene, delete dynamic data, save results
TestCompute::~TestCompute()
{
  s->Destroy();
}

//**************************************************************************************************
//Test run
void TestCompute::Run()
{
  Test::Run();
  if(!draw_scene)
		return;   //don't draw scene if has ended or splashscreen is present

  /////////////////////////////////////////////////////////////////////////////////////////////////
  //switch according to test type
  if(type == COMPUTE_NBODY)
	{
		//calculate new body positions through OpenCL
		nbody->Run();

		body = 0;
		for(int i=0; i<halfbody; i++)
		{
			sprintf(objname,"1_star%d",i);
			s->MoveObjAbs(objname,nbody->GetPos(body*4 + 0), nbody->GetPos(body*4 + 1), nbody->GetPos(body*4 + 2));
			body++;
		}
		for(int i=0; i<halfbody; i++)
		{
			sprintf(objname,"2_star%d",i);
			s->MoveObjAbs(objname,nbody->GetPos(body*4 + 0), nbody->GetPos(body*4 + 1), nbody->GetPos(body*4 + 2));
			body++;
		}

		s->RotateCameraAbs(anim*5.0f, A_Y);
		s->Redraw();
		msg = "Perf: " + num2str(nbody->GetFLOPS()) + " GFLOPS";
		s->DrawScreenText(msg.c_str(),90.0,1.0);
	}
}

//*************************************************************************************************
//Scene animation (every 10ms)
void TestCompute::Animate()
{
  Test::Animate();
}

//**************************************************************************************************
//Save results to XML file
void TestCompute::SaveResults()
{
  Test::SaveResults();

	//performance in GFLOP/s

	//final score
	test_score += nbody->GetTotalFLOPS()/total_frames * 15;
	cout<<"SCORE: "<<test_score<<endl;
  results<<"<compute>"
				"<"<<testname<<">\n"
        "<fps>"<<avg_fps<<"</fps>\n"
				"<GFLOPS>"<<nbody->GetTotalFLOPS()/total_frames<<"</GFLOPS>\n"
				"<score>"<<test_score<<"</score>"
        "</"<<testname<<">\n"
				"</compute>\n";
}
#endif

